"""
Estimation of power systems weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_POWER_SYSTEMS_MASS


@RegisterSubmodel(SERVICE_POWER_SYSTEMS_MASS, "fastoad.submodel.weight.mass.systems.power.legacy")
class PowerSystemsWeight(om.ExplicitComponent):
    """
    Weight estimation for power systems (generation and distribution)

    This includes:

    - Auxiliary Power Unit (APU)
    - electric systems
    - hydraulic systems

    Based on formulas in :cite:`supaero:2014`, mass contribution C1
    """

    def setup(self):
        self.add_input("data:geometry:cabin:NPAX1", val=np.nan)
        self.add_input("data:weight:airframe:flight_controls:mass", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:MTOW", val=np.nan, units="kg")
        self.add_input("tuning:weight:systems:power:auxiliary_power_unit:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:power:auxiliary_power_unit:mass:offset", val=0.0, units="kg"
        )
        self.add_input("tuning:weight:systems:power:electric_systems:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:power:electric_systems:mass:offset", val=0.0, units="kg"
        )
        self.add_input("tuning:weight:systems:power:hydraulic_systems:mass:k", val=1.0)
        self.add_input(
            "tuning:weight:systems:power:hydraulic_systems:mass:offset", val=0.0, units="kg"
        )
        self.add_input("settings:weight:systems:power:mass:k_elec", val=1.0)

        self.add_output("data:weight:systems:power:auxiliary_power_unit:mass", units="kg")
        self.add_output("data:weight:systems:power:electric_systems:mass", units="kg")
        self.add_output("data:weight:systems:power:hydraulic_systems:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    # pylint: disable=too-many-locals
    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        npax1 = inputs["data:geometry:cabin:NPAX1"]
        flight_controls_weight = inputs["data:weight:airframe:flight_controls:mass"]
        mtow = inputs["data:weight:aircraft:MTOW"]
        k_c11 = inputs["tuning:weight:systems:power:auxiliary_power_unit:mass:k"]
        offset_c11 = inputs["tuning:weight:systems:power:auxiliary_power_unit:mass:offset"]
        k_c12 = inputs["tuning:weight:systems:power:electric_systems:mass:k"]
        offset_c12 = inputs["tuning:weight:systems:power:electric_systems:mass:offset"]
        k_c13 = inputs["tuning:weight:systems:power:hydraulic_systems:mass:k"]
        offset_c13 = inputs["tuning:weight:systems:power:hydraulic_systems:mass:offset"]
        k_elec = inputs["settings:weight:systems:power:mass:k_elec"]

        # Mass of auxiliary power unit
        temp_c11 = 11.3 * npax1**0.64
        outputs["data:weight:systems:power:auxiliary_power_unit:mass"] = (
            k_c11 * temp_c11 + offset_c11
        )

        # Mass of electric system
        temp_c12 = k_elec * (0.444 * mtow**0.66 + 2.54 * npax1 + 0.254 * flight_controls_weight)
        outputs["data:weight:systems:power:electric_systems:mass"] = k_c12 * temp_c12 + offset_c12

        # Mass of the hydraulic system
        temp_c13 = k_elec * (0.256 * mtow**0.66 + 1.46 * npax1 + 0.146 * flight_controls_weight)
        outputs["data:weight:systems:power:hydraulic_systems:mass"] = k_c13 * temp_c13 + offset_c13



"""Computation of trim drag."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_CD_TRIM


@RegisterSubmodel(SERVICE_CD_TRIM, "fastoad.submodel.aerodynamics.CD.trim.legacy")
class CdTrim(om.ExplicitComponent):
    """Computation of trim drag."""

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        if self.options["low_speed_aero"]:
            self.add_input(
                "data:aerodynamics:aircraft:low_speed:CL",
                shape_by_conn=True,
                val=np.nan,
            )
            self.add_output(
                "data:aerodynamics:aircraft:low_speed:CD:trim",
                copy_shape="data:aerodynamics:aircraft:low_speed:CL",
            )
        else:
            self.add_input("data:aerodynamics:aircraft:cruise:CL", shape_by_conn=True, val=np.nan)
            self.add_output(
                "data:aerodynamics:aircraft:cruise:CD:trim",
                copy_shape="data:aerodynamics:aircraft:cruise:CL",
            )

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        if self.options["low_speed_aero"]:
            cl = inputs["data:aerodynamics:aircraft:low_speed:CL"]
        else:
            cl = inputs["data:aerodynamics:aircraft:cruise:CL"]

        cd_trim = []

        for cl_val in cl:
            cd_trim.append(5.89 * pow(10, -4) * cl_val)

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:aircraft:low_speed:CD:trim"] = cd_trim
        else:
            outputs["data:aerodynamics:aircraft:cruise:CD:trim"] = cd_trim



"""
Estimation of wing ToC
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math

import fastoad.api as oad
import numpy as np
import openmdao.api as om

from ..constants import SERVICE_WING_GEOMETRY_THICKNESS


@oad.RegisterSubmodel(
    SERVICE_WING_GEOMETRY_THICKNESS, "fastoad.submodel.geometry.wing.thickness.legacy"
)
class ComputeToCWing(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Wing ToC estimation"""

    def setup(self):
        self.add_input("data:TLAR:cruise_mach", val=np.nan)
        self.add_input("data:geometry:wing:sweep_25", val=np.nan, units="deg")
        self.add_input("data:geometry:wing:kink:span_ratio", val=np.nan)

        self.add_output("data:geometry:wing:thickness_ratio")
        self.add_output("data:geometry:wing:root:thickness_ratio")
        self.add_output("data:geometry:wing:kink:thickness_ratio")
        self.add_output("data:geometry:wing:tip:thickness_ratio")

    def setup_partials(self):
        self.declare_partials("data:geometry:wing:thickness_ratio", "*", method="fd")
        self.declare_partials("data:geometry:wing:root:thickness_ratio", "*", method="fd")
        self.declare_partials("data:geometry:wing:kink:thickness_ratio", "*", method="fd")
        self.declare_partials("data:geometry:wing:tip:thickness_ratio", "*", method="fd")

    def compute(self, inputs, outputs):
        cruise_mach = inputs["data:TLAR:cruise_mach"]
        sweep_25 = inputs["data:geometry:wing:sweep_25"]
        wing_break = inputs["data:geometry:wing:kink:span_ratio"]

        # Relative thickness
        el_aero = 0.89 - (cruise_mach + 0.02) * math.sqrt(math.cos(sweep_25 / 180.0 * math.pi))
        el_emp = 1.24 * el_aero
        if wing_break == 0.0:
            el_break = el_emp  # Kink is set on root chord
        else:
            el_break = 0.94 * el_aero
        el_ext = 0.86 * el_aero

        outputs["data:geometry:wing:thickness_ratio"] = el_aero
        outputs["data:geometry:wing:root:thickness_ratio"] = el_emp
        outputs["data:geometry:wing:kink:thickness_ratio"] = el_break
        outputs["data:geometry:wing:tip:thickness_ratio"] = el_ext



"""
Estimation of vertical tail chords and span
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

import openmdao.api as om


class ComputeVTChords(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Vertical tail chords and span estimation"""

    def setup(self):
        self.add_input("data:geometry:vertical_tail:aspect_ratio", val=np.nan)
        self.add_input("data:geometry:vertical_tail:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:vertical_tail:taper_ratio", val=np.nan)

        self.add_output("data:geometry:vertical_tail:span", units="m")
        self.add_output("data:geometry:vertical_tail:root:chord", units="m")
        self.add_output("data:geometry:vertical_tail:tip:chord", units="m")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:vertical_tail:span",
            ["data:geometry:vertical_tail:aspect_ratio", "data:geometry:vertical_tail:area"],
            method="fd",
        )
        self.declare_partials("data:geometry:vertical_tail:root:chord", "*", method="fd")
        self.declare_partials("data:geometry:vertical_tail:tip:chord", "*", method="fd")

    def compute(self, inputs, outputs):
        lambda_vt = inputs["data:geometry:vertical_tail:aspect_ratio"]
        s_v = inputs["data:geometry:vertical_tail:area"]
        taper_v = inputs["data:geometry:vertical_tail:taper_ratio"]

        b_v = np.sqrt(max(lambda_vt * s_v, 0.1))  #
        root_chord = s_v * 2 / (1 + taper_v) / b_v
        tip_chord = root_chord * taper_v

        outputs["data:geometry:vertical_tail:span"] = b_v
        outputs["data:geometry:vertical_tail:root:chord"] = root_chord
        outputs["data:geometry:vertical_tail:tip:chord"] = tip_chord



"""
Management of 2D wing profiles
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import operator
from collections import namedtuple
from typing import Sequence, Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

Coordinates2D = namedtuple("Coordinates2D", ["x", "y"])

X = "x"
Z = "z"
THICKNESS = "thickness"


class Profile:
    """Class for managing 2D wing profiles
    :param chord_length:
    """

    # pylint: disable=invalid-name  # X and Z are valid names in this context

    def __init__(self, chord_length: float = 0.0):
        self._rel_mean_line_and_thickness = pd.DataFrame(columns=[X, Z, THICKNESS])
        """
        Data of mean line and thickness, computed after inputs of :meth:`set_points`_.

        DataFrame keys are 'x', 'z' and 'thickness'.
        - 'x' and 'z' are relative to chord_length
        - 'thickness' is relative to max thickness (and given according to 'x')
        """

        self.chord_length: float = chord_length
        """ in meters """

        self._max_relative_thickness: float = 0.0
        """ max thickness / chord length"""

    @property
    def thickness_ratio(self) -> float:
        """thickness-to-chord ratio"""
        return self._max_relative_thickness

    @thickness_ratio.setter
    def thickness_ratio(self, value: float):
        # FIXME: mean line is modified accordingly to conform to legacy algorithm, but it
        #        is questionable
        if self._max_relative_thickness != 0.0:
            coeff = value / self._max_relative_thickness
            self._rel_mean_line_and_thickness[Z] *= coeff
        self._max_relative_thickness = value

    def set_points(
        self,
        x: Sequence,
        z: Sequence,
        keep_chord_length: bool = True,
        keep_relative_thickness: bool = True,
    ):
        """
        Sets points of the 2D profile.

        Provided points are expected to be in order around the profile (clockwise
        or anti-clockwise).

        :param x: in meters
        :param z: in meters
        :param keep_relative_thickness:
        :param keep_chord_length:
        """

        x = np.asarray(x)

        # Separate upper surface from lower surface (easier for computation
        # of thickness and mean line)
        upper, lower = self._create_upper_lower_sides(x, z)

        # Upper and lower sides are defined, we can compute mean line and thickness
        chord_length, max_thickness = self._compute_mean_line_and_thickness(upper, lower)

        if not keep_chord_length or self.chord_length == 0.0:
            self.chord_length = chord_length
        if not keep_relative_thickness or self.thickness_ratio == 0.0:
            self.thickness_ratio = max_thickness / chord_length

    def get_mean_line(self) -> pd.DataFrame:
        """Point set of mean line of the profile.

        DataFrame keys are 'x' and 'z', given in meters.
        """
        mean_line = self._rel_mean_line_and_thickness[[X, Z]] * self.chord_length
        return mean_line

    def get_relative_thickness(self) -> pd.DataFrame:
        """Point set of relative thickness of the profile.

        DataFrame keys are 'x' and 'thickness' and are relative to chord_length.
        'x' is from 0. to 1.
        """
        return self._rel_mean_line_and_thickness[[X, THICKNESS]] * [1.0, self.thickness_ratio]

    def get_upper_side(self) -> pd.DataFrame:
        """Point set of upper side of the profile.

        DataFrame keys are 'x' and 'z', given in meters.
        """
        return self._get_side_points(operator.add)

    def get_lower_side(self) -> pd.DataFrame:
        """Point set of lower side of the profile.

        DataFrame keys are 'x' and 'z', given in meters.
        """
        return self._get_side_points(operator.sub)

    def get_sides(self) -> pd.DataFrame:
        """Point set of the whole profile

        Points are given from trailing edge to trailing edge, starting by upper side.
        """
        return pd.concat(
            [self.get_upper_side().sort_values(by=X, ascending=False), self.get_lower_side()[1:]]
        )

    def _get_side_points(self, operator_) -> pd.DataFrame:
        """
        Computes upper or lower side points.

        operator_ ==  operator.add() -> upper side
        operator_ ==  operator.sub() -> lower side
        """
        mean_line = self._rel_mean_line_and_thickness[[X, Z]]
        half_thickness = pd.DataFrame().reindex_like(mean_line)
        half_thickness[X] = 0.0
        half_thickness[Z] = (
            self._rel_mean_line_and_thickness[THICKNESS] / 2.0 * self.thickness_ratio
        )
        points = operator_(mean_line, half_thickness) * self.chord_length
        return points

    def _compute_mean_line_and_thickness(
        self, upper_side_points, lower_side_points
    ) -> Tuple[float, float]:
        """
        Computes mean line and thickness from upper_side_points and lower_side_points.

        Fills self._rel_mean_line_and_thickness with relative values.
        Returns actual chord length and maximum thickness (in meters)
        """
        x = (
            pd.concat([lower_side_points[X], upper_side_points[X]])
            .sort_values()
            .drop_duplicates()
            .reset_index(drop=True)
        )

        interp_lower = interp1d(lower_side_points[X], lower_side_points[Z], kind="quadratic")
        interp_upper = interp1d(upper_side_points[X], upper_side_points[Z], kind="quadratic")
        z_sides = pd.DataFrame({"z_lower": interp_lower(x), "z_upper": interp_upper(x)})
        z = z_sides.mean(axis=1)
        thickness = z_sides.diff(axis=1).iloc[:, -1]

        chord_length = np.max(x) - np.min(x)
        max_thickness = np.max(thickness)
        self._rel_mean_line_and_thickness = pd.DataFrame(
            {X: x / chord_length, Z: z / chord_length, THICKNESS: thickness / max_thickness}
        )
        return chord_length, max_thickness

    @staticmethod
    def _create_upper_lower_sides(x: Sequence, z: Sequence) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """returns upper side points and lower side points using provided x and z"""
        # FIXME: leading and trailing edges are located roughly.
        i_leading_edge = np.argmin(x)
        i_trailing_edge = np.argmax(x)

        i1 = min(i_leading_edge, i_trailing_edge)
        i2 = max(i_leading_edge, i_trailing_edge)
        side1 = pd.DataFrame({X: x[i1 : i2 + 1], Z: z[i1 : i2 + 1]})
        side2_1 = pd.DataFrame({X: x[i2:], Z: z[i2:]})
        side2_2 = pd.DataFrame({X: x[: i1 + 1], Z: z[: i1 + 1]})
        side2 = pd.concat((side2_1, side2_2)).reset_index(drop=True)

        side1.sort_values(by=X, inplace=True)
        side2.sort_values(by=X, inplace=True)

        # At this point, side2 and side1 have the same last point, but in in case of thick
        # trailing edge, it could lead to side2 having 2 points for the same X, which will be
        # harmful in next operations.
        # In that case, we simply have to remove last point of side2, as it actually belongs to
        # side1.
        if side2[X].iloc[-1] == side2[X].iloc[-2]:
            side2 = side2.iloc[:-1]

        if np.max(side1[Z]) > np.max(side2[Z]):
            upper_side_points = side1
            lower_side_points = side2
        else:
            upper_side_points = side2
            lower_side_points = side1

        upper_side_points.drop_duplicates(inplace=True)
        lower_side_points.drop_duplicates(inplace=True)
        upper_side_points.reset_index(drop=True, inplace=True)
        lower_side_points.reset_index(drop=True, inplace=True)

        return upper_side_points, lower_side_points



"""
Estimation of wing chords (l1 and l4)
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math

import numpy as np
import openmdao.api as om


class ComputeL1AndL4Wing(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Wing chords (l1 and l4) estimation"""

    def setup(self):
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:virtual_taper_ratio", val=np.nan)
        self.add_input("data:geometry:wing:sweep_25", val=np.nan, units="rad")
        self.add_input("data:geometry:wing:sweep_100_inner", val=0.0, units="rad")

        self.add_output("data:geometry:wing:root:virtual_chord", units="m")
        self.add_output("data:geometry:wing:tip:chord", units="m")

    def setup_partials(self):
        self.declare_partials("data:geometry:wing:root:virtual_chord", "*", method="fd")
        self.declare_partials("data:geometry:wing:tip:chord", "*", method="fd")

    def compute(self, inputs, outputs):
        wing_area = inputs["data:geometry:wing:area"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        y3_wing = inputs["data:geometry:wing:kink:y"]
        y4_wing = inputs["data:geometry:wing:tip:y"]
        span = 2 * y4_wing
        sweep_25 = inputs["data:geometry:wing:sweep_25"]
        sweep_100 = inputs["data:geometry:wing:sweep_100_inner"]
        virtual_taper_ratio = inputs["data:geometry:wing:virtual_taper_ratio"]

        l1_wing = (
            wing_area
            - (y3_wing - y2_wing) * (y3_wing + y2_wing) * (math.tan(sweep_25) - math.tan(sweep_100))
        ) / (
            (1.0 + virtual_taper_ratio) / 2.0 * (span - 2 * y2_wing)
            + 2 * y2_wing
            - (3.0 * (1.0 - virtual_taper_ratio) * (y3_wing - y2_wing) * (y3_wing + y2_wing))
            / (2.0 * (span - 2 * y2_wing))
        )

        l4_wing = l1_wing * virtual_taper_ratio

        outputs["data:geometry:wing:root:virtual_chord"] = l1_wing
        outputs["data:geometry:wing:tip:chord"] = l4_wing



"""
Estimation of horizontal tail area
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from scipy.constants import g

from stdatm import Atmosphere


class ComputeHTArea(om.ExplicitComponent):
    """
    Computes area of horizontal tail plane

    Area is computed to fulfill aircraft balance requirement at rotation speed
    """

    def setup(self):
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:has_T_tail", val=np.nan)
        self.add_input("data:geometry:vertical_tail:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:vertical_tail:tip:leading_edge:x", val=np.nan, units="m")
        self.add_input(
            "data:geometry:horizontal_tail:MAC:at25percent:x:local", val=np.nan, units="m"
        )
        self.add_input("data:weight:airframe:landing_gear:main:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:airframe:landing_gear:front:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:aircraft:MTOW", val=np.nan, units="kg")
        self.add_input("settings:weight:aircraft:CG:range", val=0.3)
        self.add_input("settings:weight:airframe:landing_gear:front:weight_ratio", val=0.08)
        self.add_input(
            "settings:geometry:horizontal_tail:position_ratio_on_fuselage",
            val=0.91,
            desc="(does not apply for T-tails) distance to aircraft nose of 25% MAC of "
            "horizontal tail divided by fuselage length",
        )
        self.add_input(
            "settings:geometry:horizontal_tail:position_ratio_on_VTP",
            val=0.15,
            desc="(applies only for T-tails) distance between HTP root leading"
            "edge and VTP tip leading edge divided by VTP tip chord",
        )

        self.add_output(
            "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25", units="m", ref=30.0
        )
        self.add_output("data:geometry:horizontal_tail:wetted_area", units="m**2", ref=100.0)
        self.add_output("data:geometry:horizontal_tail:area", units="m**2", ref=50.0)

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")
        self.declare_partials(
            "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25",
            ["data:geometry:fuselage:length", "data:geometry:wing:MAC:at25percent:x"],
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        # Area of HTP is computed so its "lift" can counter the moment of weight
        # on front landing gear w.r.t. main landing gear when the CG is in its
        # most front position.

        tail_type = np.round(inputs["data:geometry:has_T_tail"])
        fuselage_length = inputs["data:geometry:fuselage:length"]
        x_wing_aero_center = inputs["data:geometry:wing:MAC:at25percent:x"]
        x_main_lg = inputs["data:weight:airframe:landing_gear:main:CG:x"]
        x_front_lg = inputs["data:weight:airframe:landing_gear:front:CG:x"]
        mtow = inputs["data:weight:aircraft:MTOW"]
        wing_area = inputs["data:geometry:wing:area"]
        wing_mac = inputs["data:geometry:wing:MAC:length"]
        cg_range = inputs["settings:weight:aircraft:CG:range"]
        front_lg_weight_ratio = inputs["settings:weight:airframe:landing_gear:front:weight_ratio"]
        htp_aero_center_ratio = inputs[
            "settings:geometry:horizontal_tail:position_ratio_on_fuselage"
        ]

        vtp_tip_chord = inputs["data:geometry:vertical_tail:tip:chord"]
        vt_le_x = inputs["data:geometry:vertical_tail:tip:leading_edge:x"]
        htp_le_position_ratio = inputs["settings:geometry:horizontal_tail:position_ratio_on_VTP"]
        htp_x0 = inputs["data:geometry:horizontal_tail:MAC:at25percent:x:local"]

        delta_lg = x_main_lg - x_front_lg
        atm = Atmosphere(0.0)
        rho = atm.density
        vspeed = atm.speed_of_sound * 0.2  # assume the corresponding Mach of VR is 0.2

        # Proportion of weight on front landing gear is equal to distance between
        # main landing gear and center of gravity, divided by distance between landing gears.

        # If CG is in the most front position, the distance between main landing gear
        # and center of gravity is:
        distance_cg_to_mlg = front_lg_weight_ratio * delta_lg + wing_mac * cg_range

        # So with this front CG, moment of (weight on front landing gear) w.r.t.
        # main landing gear is:
        m_front_lg = mtow * g * distance_cg_to_mlg

        # Moment coefficient
        pdyn = 0.5 * rho * vspeed**2
        cm_front_lg = m_front_lg / (pdyn * wing_area * wing_mac)

        # # CM of MTOW on main landing gear w.r.t 25% wing MAC
        # lever_arm = front_lg_weight_ratio * delta_lg  # lever arm wrt CoG
        # lever_arm += wing_mac * cg_range  # and now wrt 25% wing MAC
        # cm_wheel = mtow * g * lever_arm / (pdyn * wing_area * wing_mac)

        ht_volume_coeff = cm_front_lg
        if tail_type == 1:
            aero_centers_distance = (
                vt_le_x + htp_le_position_ratio * vtp_tip_chord + htp_x0 - x_wing_aero_center
            )
            wet_area_coeff = 2.0  # TODO: explore more thoroughly this coefficient
        elif tail_type == 0:
            aero_centers_distance = htp_aero_center_ratio * fuselage_length - x_wing_aero_center
            wet_area_coeff = 2.0  # TODO: explore more thoroughly this coefficient
        else:
            raise ValueError("Value of data:geometry:has_T_tail can only be 0 or 1")

        htp_area = ht_volume_coeff / aero_centers_distance * wing_area * wing_mac
        wet_area_htp = wet_area_coeff * htp_area

        outputs["data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25"] = (
            aero_centers_distance
        )
        outputs["data:geometry:horizontal_tail:wetted_area"] = wet_area_htp
        outputs["data:geometry:horizontal_tail:area"] = htp_area



"""
Estimation of yawing moment due to sideslip
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import math

import fastoad.api as oad
import numpy as np
import openmdao.api as om

from ...constants import SERVICE_FUSELAGE_CNBETA


# TODO: This belongs more to aerodynamics than geometry
@oad.RegisterSubmodel(SERVICE_FUSELAGE_CNBETA, "fastoad.submodel.geometry.fuselage.cnbeta.legacy")
class ComputeCnBetaFuselage(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Yawing moment due to sideslip estimation"""

    def setup(self):
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_height", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:front_length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:rear_length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")

        self.add_output("data:aerodynamics:fuselage:cruise:CnBeta")

    def setup_partials(self):
        self.declare_partials("data:aerodynamics:fuselage:cruise:CnBeta", "*", method="fd")

    def compute(self, inputs, outputs):
        fus_length = inputs["data:geometry:fuselage:length"]
        lav = inputs["data:geometry:fuselage:front_length"]
        lar = inputs["data:geometry:fuselage:rear_length"]
        width_max = inputs["data:geometry:fuselage:maximum_width"]
        height_max = inputs["data:geometry:fuselage:maximum_height"]
        wing_area = inputs["data:geometry:wing:area"]
        span = inputs["data:geometry:wing:span"]

        l_f = math.sqrt(width_max * height_max)
        l_cyc = fus_length - lav - lar
        # estimation of fuselage volume
        volume_fus = math.pi * l_f**2 / 4 * (0.7 * lav + 0.5 * lar + l_cyc)
        # equation from raymer book eqn. 16.47
        cn_beta_fus = -1.3 * volume_fus / wing_area / span * (l_f / width_max)

        outputs["data:aerodynamics:fuselage:cruise:CnBeta"] = cn_beta_fus



"""
FAST - Copyright (c) 2016 ONERA ISAE
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import (
    SERVICE_AIRCRAFT_CG,
    SERVICE_FLIGHT_CONTROLS_CG,
    SERVICE_GLOBAL_CG,
    SERVICE_HORIZONTAL_TAIL_CG,
    SERVICE_MLG_CG,
    SERVICE_OTHERS_CG,
    SERVICE_TANKS_CG,
    SERVICE_VERTICAL_TAIL_CG,
    SERVICE_WING_CG,
)
from ..constants import SERVICE_CENTERS_OF_GRAVITY


@RegisterSubmodel(SERVICE_CENTERS_OF_GRAVITY, "fastoad.submodel.weight.cg.legacy")
class CG(
    oad.CycleGroup,
    default_linear_solver="om.LinearBlockGS",
    default_linear_options={"iprint": 0},
    default_nonlinear_options={"maxiter": 200, "iprint": 0},
):
    """Model that computes the global center of gravity"""

    def setup(self):
        super().setup()
        self.add_subsystem(
            "ht_cg", RegisterSubmodel.get_submodel(SERVICE_HORIZONTAL_TAIL_CG), promotes=["*"]
        )
        self.add_subsystem(
            "vt_cg", RegisterSubmodel.get_submodel(SERVICE_VERTICAL_TAIL_CG), promotes=["*"]
        )
        self.add_subsystem(
            "compute_cg_wing", RegisterSubmodel.get_submodel(SERVICE_WING_CG), promotes=["*"]
        )
        self.add_subsystem(
            "compute_cg_control_surface",
            RegisterSubmodel.get_submodel(SERVICE_FLIGHT_CONTROLS_CG),
            promotes=["*"],
        )
        self.add_subsystem(
            "compute_cg_tanks", RegisterSubmodel.get_submodel(SERVICE_TANKS_CG), promotes=["*"]
        )
        self.add_subsystem(
            "compute_cg_others", RegisterSubmodel.get_submodel(SERVICE_OTHERS_CG), promotes=["*"]
        )
        self.add_subsystem(
            "compute_cg", RegisterSubmodel.get_submodel(SERVICE_GLOBAL_CG), promotes=["*"]
        )
        self.add_subsystem(
            "update_mlg", RegisterSubmodel.get_submodel(SERVICE_MLG_CG), promotes=["*"]
        )
        self.add_subsystem(
            "aircraft", RegisterSubmodel.get_submodel(SERVICE_AIRCRAFT_CG), promotes=["*"]
        )


@RegisterSubmodel(SERVICE_AIRCRAFT_CG, "fastoad.submodel.weight.cg.aircraft.legacy")
class ComputeAircraftCG(om.ExplicitComponent):
    """Compute position of aircraft CG from CG ratio"""

    def setup(self):
        self.add_input("data:weight:aircraft:CG:aft:MAC_position", val=np.nan)
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")

        self.add_output("data:weight:aircraft:CG:aft:x", units="m")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        cg_ratio = inputs["data:weight:aircraft:CG:aft:MAC_position"]
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        mac_position = inputs["data:geometry:wing:MAC:at25percent:x"]

        outputs["data:weight:aircraft:CG:aft:x"] = (
            mac_position - 0.25 * l0_wing + cg_ratio * l0_wing
        )



"""Initialization of CL vector."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import POLAR_POINT_COUNT, SERVICE_INITIALIZE_CL


@RegisterSubmodel(SERVICE_INITIALIZE_CL, "fastoad.submodel.aerodynamics.initialize_CL.legacy")
class InitializeClPolar(om.ExplicitComponent):
    """Initialization of CL vector."""

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.add_input("tuning:aerodynamics:aircraft:cruise:CL:k", val=np.nan)
        self.add_input("tuning:aerodynamics:aircraft:cruise:CL:offset", val=np.nan)
        self.add_input("tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:k", val=np.nan)
        self.add_input("tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:offset", val=np.nan)

        if self.options["low_speed_aero"]:
            self.add_output("data:aerodynamics:aircraft:low_speed:CL", shape=POLAR_POINT_COUNT)
        else:
            self.add_output("data:aerodynamics:aircraft:cruise:CL", shape=POLAR_POINT_COUNT)

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        k_cl = inputs["tuning:aerodynamics:aircraft:cruise:CL:k"]
        offset_cl = inputs["tuning:aerodynamics:aircraft:cruise:CL:offset"]
        k_winglet_cl = inputs["tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:k"]
        offset_winglet_cl = inputs["tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:offset"]

        # FIXME: initialization of CL range should be done more directly, without these coefficients
        cl = np.arange(0.0, 1.5, 0.01) * k_cl * k_winglet_cl + offset_cl + offset_winglet_cl

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:aircraft:low_speed:CL"] = cl
        else:
            outputs["data:aerodynamics:aircraft:cruise:CL"] = cl



"""
Estimation of wing B50
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math

import numpy as np
import openmdao.api as om


class ComputeB50(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Wing B50 estimation"""

    def setup(self):
        self.add_input("data:geometry:wing:tip:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:virtual_chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")

        self.add_output("data:geometry:wing:b_50", units="m")

    def setup_partials(self):
        self.declare_partials("data:geometry:wing:b_50", "*", method="fd")

    def compute(self, inputs, outputs):
        x4_wing = inputs["data:geometry:wing:tip:leading_edge:x:local"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        y4_wing = inputs["data:geometry:wing:tip:y"]
        l1_wing = inputs["data:geometry:wing:root:virtual_chord"]
        l4_wing = inputs["data:geometry:wing:tip:chord"]
        span = inputs["data:geometry:wing:span"]

        sweep_50 = math.atan((x4_wing + l4_wing * 0.5 - 0.5 * l1_wing) / (y4_wing - y2_wing))
        b_50 = span / math.cos(sweep_50)

        outputs["data:geometry:wing:b_50"] = b_50



"""Main components for mass breakdown."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import (
    SERVICE_AIRFRAME_MASS,
    SERVICE_CREW_MASS,
    SERVICE_FURNITURE_MASS,
    SERVICE_OWE,
    SERVICE_PAYLOAD_MASS,
    SERVICE_PROPULSION_MASS,
    SERVICE_SYSTEMS_MASS,
)
from .update_mlw_and_mzfw import UpdateMLWandMZFW
from ..constants import SERVICE_MASS_BREAKDOWN
from ...constants import PAYLOAD_FROM_NPAX


@RegisterSubmodel(SERVICE_MASS_BREAKDOWN, "fastoad.submodel.weight.mass.legacy")
class MassBreakdown(
    oad.CycleGroup,
    default_linear_solver="om.LinearBlockGS",
    default_linear_options={"iprint": 0},
    default_nonlinear_options={"maxiter": 50, "iprint": 0},
):
    """
    Computes analytically the mass of each part of the aircraft, and the resulting sum,
    the Overall Weight Empty (OWE).

    Some models depend on MZFW (Max Zero Fuel Weight), MLW (Max Landing Weight) and
    MTOW (Max TakeOff Weight), which depend on OWE.

    This model cycles for having consistent OWE, MZFW and MLW.

    Options:
    - payload_from_npax: If True (default), payload masses will be computed from NPAX, if False
                         design payload mass and maximum payload mass must be provided.
    """

    def initialize(self):
        super().initialize()
        self.options.declare(PAYLOAD_FROM_NPAX, types=bool, default=True)

    def setup(self):
        super().setup()
        if self.options[PAYLOAD_FROM_NPAX]:
            self.add_subsystem(
                "payload",
                RegisterSubmodel.get_submodel(SERVICE_PAYLOAD_MASS),
                promotes=["*"],
            )
        self.add_subsystem("owe", RegisterSubmodel.get_submodel(SERVICE_OWE), promotes=["*"])
        self.add_subsystem("update_mzfw_and_mlw", UpdateMLWandMZFW(), promotes=["*"])


@RegisterSubmodel(SERVICE_OWE, "fastoad.submodel.weight.mass.owe.legacy")
class OperatingWeightEmpty(om.Group):
    """Operating Empty Weight (OEW) estimation.

    This group aggregates weight from all components of the aircraft.
    """

    def setup(self):
        # Propulsion should be done before airframe, because it drives pylon mass.
        self.add_subsystem(
            "propulsion_weight",
            RegisterSubmodel.get_submodel(SERVICE_PROPULSION_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "airframe_weight",
            RegisterSubmodel.get_submodel(SERVICE_AIRFRAME_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "systems_weight",
            RegisterSubmodel.get_submodel(SERVICE_SYSTEMS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "furniture_weight",
            RegisterSubmodel.get_submodel(SERVICE_FURNITURE_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "crew_weight",
            RegisterSubmodel.get_submodel(SERVICE_CREW_MASS),
            promotes=["*"],
        )

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:aircraft:OWE",
            [
                "data:weight:airframe:mass",
                "data:weight:propulsion:mass",
                "data:weight:systems:mass",
                "data:weight:furniture:mass",
                "data:weight:crew:mass",
            ],
            units="kg",
            desc="Mass of crew",
        )

        self.add_subsystem("OWE_sum", weight_sum, promotes=["*"])



"""Constants for geometry submodels."""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

SERVICE_FUSELAGE_GEOMETRY_BASIC = "service.geometry.fuselage.basic"
SERVICE_FUSELAGE_GEOMETRY_WITH_CABIN_SIZING = "service.geometry.fuselage.with_cabin_sizing"
SERVICE_FUSELAGE_CNBETA = "service.geometry.fuselage.cnbeta"
SERVICE_HORIZONTAL_TAIL_GEOMETRY = "service.geometry.horizontal_tail"
SERVICE_NACELLE_PYLON_GEOMETRY = "service.geometry.nacelle_and_pylon"
SERVICE_VERTICAL_TAIL_GEOMETRY = "service.geometry.vertical_tail"
SERVICE_WING_GEOMETRY = "service.geometry.wing"
SERVICE_AIRCRAFT_WETTED_AREA = "service.geometry.aircraft.wetted_area"
SERVICE_AIRCRAFT_AERODYNAMIC_CENTER = "service.geometry.aircraft.aerodynamic_center"



"""
Computation of wing area
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.constants import ModelDomain
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem, RegisterSubmodel

from .constants import (
    SERVICE_WING_AREA_CONSTRAINT_AERO,
    SERVICE_WING_AREA_CONSTRAINT_GEOM,
    SERVICE_WING_AREA_LOOP_AERO,
    SERVICE_WING_AREA_LOOP_GEOM,
)


@RegisterOpenMDAOSystem("fastoad.loop.wing_area", domain=ModelDomain.OTHER)
class ComputeWingArea(om.Group):
    """
    Computes needed wing area for:
      - having enough lift at required approach speed
      - being able to load enough fuel to achieve the sizing mission
    """

    def initialize(self):
        super().initialize()
        self.options.declare("use_fuel", default=True, types=bool)
        self.options.declare("use_approach_speed", default=True, types=bool)

    def setup(self):
        if self.options["use_fuel"]:
            self.add_subsystem(
                "wing_area_geom",
                RegisterSubmodel.get_submodel(SERVICE_WING_AREA_LOOP_GEOM),
                promotes_inputs=["*"],
                promotes_outputs=[],
            )
            self.connect("wing_area_geom.wing_area:geom", "wing_area.wing_area:geom")

        if self.options["use_approach_speed"]:
            self.add_subsystem(
                "wing_area_aero",
                RegisterSubmodel.get_submodel(SERVICE_WING_AREA_LOOP_AERO),
                promotes_inputs=["*"],
                promotes_outputs=[],
            )
            self.connect("wing_area_aero.wing_area:aero", "wing_area.wing_area:aero")

        if self.options["use_fuel"] or self.options["use_approach_speed"]:
            self.add_subsystem(
                "wing_area",
                _ComputeWingArea(),
                promotes_inputs=[],
                promotes_outputs=["*"],
            )
        self.add_subsystem(
            "geom_constraint",
            RegisterSubmodel.get_submodel(SERVICE_WING_AREA_CONSTRAINT_GEOM),
            promotes=["*"],
        )
        self.add_subsystem(
            "aero_constraint",
            RegisterSubmodel.get_submodel(SERVICE_WING_AREA_CONSTRAINT_AERO),
            promotes=["*"],
        )


class _ComputeWingArea(om.ExplicitComponent):
    """
    Computation of wing area from needed approach speed and mission fuel and taking whichever
    is greatest
    """

    def setup(self):
        self.add_input("wing_area:aero", units="m**2", val=0.0)
        self.add_input("wing_area:geom", units="m**2", val=0.0)

        self.add_output("data:geometry:wing:area", val=100.0, units="m**2")

        self.declare_partials(of="*", wrt="*", method="exact")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        wing_area_mission = inputs["wing_area:geom"]
        wing_area_approach = inputs["wing_area:aero"]

        outputs["data:geometry:wing:area"] = np.nanmax([wing_area_mission, wing_area_approach])

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        wing_area_mission = inputs["wing_area:geom"]
        wing_area_approach = inputs["wing_area:aero"]

        partials["data:geometry:wing:area", "wing_area:geom"] = (
            1.0 if wing_area_mission > wing_area_approach else 0.0
        )
        partials["data:geometry:wing:area", "wing_area:aero"] = (
            0.0 if wing_area_mission > wing_area_approach else 1.0
        )



"""
Estimation of geometry of vertical tail
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .components import (
    ComputeVTChords,
    ComputeVTClalpha,
    ComputeVTDistance,
    ComputeVTLocalPositions,
    ComputeVTMAC,
    ComputeVTSweep,
)
from .vt_global_positions import VTChordGlobalPositions
from ...constants import (
    SERVICE_FUSELAGE_CNBETA,
    SERVICE_VERTICAL_TAIL_GEOMETRY,
)


@oad.RegisterSubmodel(
    SERVICE_VERTICAL_TAIL_GEOMETRY, "fastoad.submodel.geometry.vertical_tail.legacy"
)
class ComputeVerticalTailGeometry(om.Group):
    """Vertical tail geometry estimation"""

    def setup(self):
        self.add_subsystem(
            "fuselage_cnbeta",
            RegisterSubmodel.get_submodel(SERVICE_FUSELAGE_CNBETA),
            promotes=["*"],
        )
        self.add_subsystem("vt_aspect_ratio", ComputeVTDistance(), promotes=["*"])
        self.add_subsystem("vt_clalpha", ComputeVTClalpha(), promotes=["*"])
        self.add_subsystem("vt_chords", ComputeVTChords(), promotes=["*"])
        self.add_subsystem("vt_mac", ComputeVTMAC(), promotes=["*"])
        self.add_subsystem("vt_sweep", ComputeVTSweep(), promotes=["*"])
        self.add_subsystem("vt_local_positions", ComputeVTLocalPositions(), promotes=["*"])
        self.add_subsystem("global_positions", VTChordGlobalPositions(), promotes=["*"])



"""
Estimation of wing mean aerodynamic chord
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om


# TODO: it would be good to have a function to compute MAC for HT, VT and WING
class ComputeMACWing(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Wing mean aerodynamic chord estimation"""

    def setup(self):
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:kink:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:chord", val=np.nan, units="m")

        self.add_output("data:geometry:wing:MAC:length", units="m")
        self.add_output("data:geometry:wing:MAC:leading_edge:x:local", units="m")
        self.add_output("data:geometry:wing:MAC:y", units="m")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:wing:MAC:length",
            [
                "data:geometry:wing:root:y",
                "data:geometry:wing:kink:y",
                "data:geometry:wing:tip:y",
                "data:geometry:wing:root:chord",
                "data:geometry:wing:kink:chord",
                "data:geometry:wing:tip:chord",
                "data:geometry:wing:area",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:wing:MAC:leading_edge:x:local",
            [
                "data:geometry:wing:kink:leading_edge:x:local",
                "data:geometry:wing:tip:leading_edge:x:local",
                "data:geometry:wing:root:y",
                "data:geometry:wing:kink:y",
                "data:geometry:wing:tip:y",
                "data:geometry:wing:root:chord",
                "data:geometry:wing:kink:chord",
                "data:geometry:wing:tip:chord",
                "data:geometry:wing:area",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:wing:MAC:y",
            [
                "data:geometry:wing:root:y",
                "data:geometry:wing:kink:y",
                "data:geometry:wing:tip:y",
                "data:geometry:wing:root:chord",
                "data:geometry:wing:kink:chord",
                "data:geometry:wing:tip:chord",
                "data:geometry:wing:area",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs):
        wing_area = inputs["data:geometry:wing:area"]
        x3_wing = inputs["data:geometry:wing:kink:leading_edge:x:local"]
        x4_wing = inputs["data:geometry:wing:tip:leading_edge:x:local"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        y3_wing = inputs["data:geometry:wing:kink:y"]
        y4_wing = inputs["data:geometry:wing:tip:y"]
        l2_wing = inputs["data:geometry:wing:root:chord"]
        l3_wing = inputs["data:geometry:wing:kink:chord"]
        l4_wing = inputs["data:geometry:wing:tip:chord"]

        l0_wing = (
            3 * y2_wing * l2_wing**2
            + (y3_wing - y2_wing) * (l2_wing**2 + l3_wing**2 + l2_wing * l3_wing)
            + (y4_wing - y3_wing) * (l3_wing**2 + l4_wing**2 + l3_wing * l4_wing)
        ) * (2 / (3 * wing_area))

        x0_wing = (
            x3_wing
            * (
                (y3_wing - y2_wing) * (2 * l3_wing + l2_wing)
                + (y4_wing - y3_wing) * (2 * l3_wing + l4_wing)
            )
            + x4_wing * (y4_wing - y3_wing) * (2 * l4_wing + l3_wing)
        ) / (3 * wing_area)

        y0_wing = (
            3 * y2_wing**2 * l2_wing
            + (y3_wing - y2_wing)
            * (l3_wing * (y2_wing + 2 * y3_wing) + l2_wing * (y3_wing + 2 * y2_wing))
            + (y4_wing - y3_wing)
            * (l4_wing * (y3_wing + 2 * y4_wing) + l3_wing * (y4_wing + 2 * y3_wing))
        ) / (3 * wing_area)

        outputs["data:geometry:wing:MAC:length"] = l0_wing
        outputs["data:geometry:wing:MAC:leading_edge:x:local"] = x0_wing
        outputs["data:geometry:wing:MAC:y"] = y0_wing



"""Computation of CL characteristics at low speed."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_ALPHA


@RegisterSubmodel(SERVICE_ALPHA, "fastoad.submodel.aerodynamics.alpha.legacy")
class ComputeAlpha(om.ExplicitComponent):
    """
    Computes a linear evolution of angle of attack from CL scale and CL gradient.

    """

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        polar_type = "low_speed" if self.options["low_speed_aero"] else "cruise"
        CL0_default = 0.2 if self.options["low_speed_aero"] else 0.1  # pylint: disable=invalid-name
        self.add_input(
            f"data:aerodynamics:aircraft:{polar_type}:CL",
            shape_by_conn=True,
            val=np.nan,
        )
        self.add_input(
            f"data:aerodynamics:aircraft:{polar_type}:CL_alpha",
            val=np.nan,
            units="1/rad",
        )
        self.add_input(f"data:aerodynamics:aircraft:{polar_type}:CL0", val=CL0_default)

        self.add_output(
            f"data:aerodynamics:aircraft:{polar_type}:AoA",
            copy_shape=f"data:aerodynamics:aircraft:{polar_type}:CL",
            units="rad",
        )

    def setup_partials(self):
        polar_type = "low_speed" if self.options["low_speed_aero"] else "cruise"
        self.declare_partials(f"data:aerodynamics:aircraft:{polar_type}:AoA", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        polar_type = "low_speed" if self.options["low_speed_aero"] else "cruise"
        CL = inputs[f"data:aerodynamics:aircraft:{polar_type}:CL"]  # pylint: disable=invalid-name
        CLalpha = inputs[  # pylint: disable=invalid-name
            f"data:aerodynamics:aircraft:{polar_type}:CL_alpha"
        ]
        CL0 = inputs[f"data:aerodynamics:aircraft:{polar_type}:CL0"]  # pylint: disable=invalid-name

        outputs[f"data:aerodynamics:aircraft:{polar_type}:AoA"] = (CL - CL0) / CLalpha



"""Parametric turbofan engine."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import math
from typing import Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from fastoad.constants import EngineSetting
from fastoad.exceptions import FastUnknownEngineSettingError
from fastoad.model_base import FlightPoint
from fastoad.model_base.propulsion import AbstractFuelPropulsion
from scipy.interpolate import interp1d
from stdatm import Atmosphere

from .constants import (
    ALPHA,
    ATM_SEA_LEVEL,
    ATM_TROPOPAUSE,
    A_FM,
    A_MS,
    BETA,
    B_FM,
    B_MS,
    C_FM,
    C_MS,
    D_FM,
    D_MS,
    E_FM,
    E_MS,
    MAX_SFC_RATIO_COEFF,
)
from .exceptions import FastRubberEngineInconsistentInputParametersError

# Logger for this module
_LOGGER = logging.getLogger(__name__)


class RubberEngine(AbstractFuelPropulsion):
    def __init__(
        self,
        bypass_ratio: float,
        overall_pressure_ratio: float,
        turbine_inlet_temperature: float,
        mto_thrust: float,
        maximum_mach: float,
        design_altitude: float,
        delta_t4_climb: float = -50,
        delta_t4_cruise: float = -100,
        k_sfc_sl: float = 1.0,
        k_sfc_cr: float = 1.0,
    ):
        """
        Parametric turbofan engine.

        It computes engine characteristics using analytical model from following
        sources:

        .. bibliography:: ../refs.bib
           :filter: docname in docnames

        :param bypass_ratio:
        :param overall_pressure_ratio:
        :param turbine_inlet_temperature: (unit=K) also noted T4
        :param mto_thrust: (unit=N) Maximum TakeOff thrust, i.e. maximum thrust
                           on ground at speed 0, also noted F0
        :param maximum_mach:
        :param design_altitude: (unit=m)
        :param delta_t4_climb: (unit=K) difference between T4 during climb and design T4
        :param delta_t4_cruise: (unit=K) difference between T4 during cruise and design T4
        :param k_sfc_sl: SFC correction at sea level and below
        :param k_sfc_cr: SFC correction at 43000ft and above in cruise
        """
        # pylint: disable=too-many-arguments  # they define the engine

        self.bypass_ratio = bypass_ratio
        self.overall_pressure_ratio = overall_pressure_ratio
        self.t_4 = turbine_inlet_temperature
        self.f_0 = mto_thrust
        self.mach_max = maximum_mach
        self.design_alt = design_altitude
        self.k_sfc_sl = k_sfc_sl
        self.k_sfc_cr = k_sfc_cr

        # This dictionary is expected to have a dT4 value for all EngineSetting values
        self.dt4_values = {
            EngineSetting.TAKEOFF: 0.0,
            EngineSetting.CLIMB: delta_t4_climb,
            EngineSetting.CRUISE: delta_t4_cruise,
            EngineSetting.IDLE: delta_t4_cruise,
        }

        # ... so check that all EngineSetting values are in dict
        unknown_keys = [key for key in EngineSetting if key not in self.dt4_values.keys()]
        if unknown_keys:
            raise FastUnknownEngineSettingError("Unknown flight phases: %s", unknown_keys)

    def compute_flight_points(self, flight_points: Union[FlightPoint, pd.DataFrame]):
        # pylint: disable=too-many-arguments  # they define the trajectory
        sfc, thrust_rate, thrust = self.compute_flight_points_from_dt4(
            flight_points.mach,
            flight_points.altitude,
            self._get_delta_t4(flight_points.engine_setting),
            flight_points.isa_offset,
            flight_points.thrust_is_regulated,
            flight_points.thrust_rate,
            flight_points.thrust,
        )
        # flight_points.sfc = sfc raises a warning if flight_points is a DataFrame that has not
        # already this field, so we add needed fields before setting values
        if isinstance(flight_points, pd.DataFrame):
            new_column_names = flight_points.columns.tolist()
            for name in ["sfc", "thrust_rate", "thrust"]:
                if name not in new_column_names:
                    flight_points.insert(len(flight_points.columns), name, value=np.nan)

        # SFC correction for NEO engines dependent on altitude.
        k_sfc_alt = interp1d(
            [-1000.0, 0.0, 13106.4, 20000.0],
            np.hstack((self.k_sfc_sl, self.k_sfc_sl, self.k_sfc_cr, self.k_sfc_cr)),
        )
        k_sfc = k_sfc_alt(flight_points.altitude)

        flight_points.sfc = sfc * k_sfc
        flight_points.thrust_rate = thrust_rate
        flight_points.thrust = thrust

    def compute_flight_points_from_dt4(
        self,
        mach: Union[float, Sequence],
        altitude: Union[float, Sequence],
        delta_t4: Union[float, Sequence],
        isa_offset: Union[float, Sequence] = 0,
        thrust_is_regulated: Optional[Union[bool, Sequence]] = None,
        thrust_rate: Optional[Union[float, Sequence]] = None,
        thrust: Optional[Union[float, Sequence]] = None,
    ) -> Tuple[Union[float, Sequence], Union[float, Sequence], Union[float, Sequence]]:
        # pylint: disable=too-many-arguments  # they define the trajectory
        """
        Same as :meth:`compute_flight_points` except that delta_t4 is used directly
        instead of specifying flight engine_setting.

        :param mach: Mach number
        :param altitude: (unit=m) altitude w.r.t. to sea level
        :param delta_t4: (unit=K) difference between operational and design values of
                         turbine inlet temperature in K
        :param isa_offset: (unit=degK) temperature difference from isa conditions
        :param thrust_is_regulated: tells if thrust_rate or thrust should be used (works
                                    element-wise)
        :param thrust_rate: thrust rate (unit=none)
        :param thrust: required thrust (unit=N)
        :return: SFC (in kg/s/N), thrust rate, thrust (in N)
        """
        mach = np.asarray(mach)
        altitude = np.asarray(altitude)
        delta_t4 = np.asarray(delta_t4)
        isa_offset = np.asarray(isa_offset)

        if thrust_is_regulated is not None:
            thrust_is_regulated = np.asarray(np.round(thrust_is_regulated, 0), dtype=bool)

        thrust_is_regulated, thrust_rate, thrust = self._check_thrust_inputs(
            thrust_is_regulated, thrust_rate, thrust
        )

        thrust_is_regulated = np.asarray(np.round(thrust_is_regulated, 0), dtype=bool)
        thrust_rate = np.asarray(thrust_rate)
        thrust = np.asarray(thrust)

        atmosphere = Atmosphere(altitude, delta_t=isa_offset, altitude_in_feet=False)

        max_thrust = self.max_thrust(atmosphere, mach, delta_t4)

        # We compute thrust values from thrust rates when needed
        idx = np.logical_not(thrust_is_regulated)
        if np.size(max_thrust) == 1:
            maximum_thrust = max_thrust
            out_thrust_rate = thrust_rate
            out_thrust = thrust
        else:
            out_thrust_rate = (
                np.full(np.shape(max_thrust), thrust_rate.item())
                if np.size(thrust_rate) == 1
                else thrust_rate
            )
            out_thrust = (
                np.full(np.shape(max_thrust), thrust.item()) if np.size(thrust) == 1 else thrust
            )

            maximum_thrust = max_thrust[idx]

        if np.any(idx):
            out_thrust[idx] = out_thrust_rate[idx] * maximum_thrust

        # thrust_rate is obtained from entire thrust vector (could be optimized if needed,
        # as some thrust rates that are computed may have been provided as input)
        out_thrust_rate = out_thrust / max_thrust

        # Now SFC can be computed
        sfc_0 = self.sfc_at_max_thrust(atmosphere, mach)
        sfc = sfc_0 * self.sfc_ratio(altitude, out_thrust_rate)

        return sfc, out_thrust_rate, out_thrust

    @staticmethod
    def _check_thrust_inputs(
        thrust_is_regulated: Optional[Union[float, Sequence]],
        thrust_rate: Optional[Union[float, Sequence]],
        thrust: Optional[Union[float, Sequence]],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Checks that inputs are consistent and return them in proper shape.

        Some of the inputs can be None, but outputs will be proper numpy arrays.

        :param thrust_is_regulated:
        :param thrust_rate:
        :param thrust:
        :return: the inputs, but transformed in numpy arrays.
        """
        # Ensure they are numpy array
        if thrust_is_regulated is not None:
            # As OpenMDAO may provide floats that could be slightly different
            # from 0. or 1., a rounding operation is needed before converting
            # to booleans
            thrust_is_regulated = np.asarray(np.round(thrust_is_regulated, 0), dtype=bool)
        if thrust_rate is not None:
            thrust_rate = np.asarray(thrust_rate)
        if thrust is not None:
            thrust = np.asarray(thrust)

        # Check inputs: if use_thrust_rate is None, we will use the provided input between
        # thrust_rate and thrust
        if thrust_is_regulated is None:
            if thrust_rate is not None:
                thrust_is_regulated = False
                thrust = np.empty_like(thrust_rate)
            elif thrust is not None:
                thrust_is_regulated = True
                thrust_rate = np.empty_like(thrust)
            else:
                raise FastRubberEngineInconsistentInputParametersError(
                    "When use_thrust_rate is None, either thrust_rate or thrust should be provided."
                )

        elif np.size(thrust_is_regulated) == 1:
            # Check inputs: if use_thrust_rate is a scalar, the matching input(thrust_rate or
            # thrust) must be provided.
            if thrust_is_regulated:
                if thrust is None:
                    raise FastRubberEngineInconsistentInputParametersError(
                        "When thrust_is_regulated is True, thrust should be provided."
                    )
                thrust_rate = np.empty_like(thrust)
            else:
                if thrust_rate is None:
                    raise FastRubberEngineInconsistentInputParametersError(
                        "When thrust_is_regulated is False, thrust_rate should be provided."
                    )
                thrust = np.empty_like(thrust_rate)

        else:
            # Check inputs: if use_thrust_rate is not a scalar, both thrust_rate and thrust must be
            # provided and have the same shape as use_thrust_rate
            if thrust_rate is None or thrust is None:
                raise FastRubberEngineInconsistentInputParametersError(
                    "When thrust_is_regulated is a sequence, both thrust_rate and thrust should be "
                    "provided."
                )
            if np.shape(thrust_rate) != np.shape(thrust_is_regulated) or np.shape(
                thrust
            ) != np.shape(thrust_is_regulated):
                raise FastRubberEngineInconsistentInputParametersError(
                    "When use_thrust_rate is a sequence, both thrust_rate and thrust should have "
                    "same shape as use_thrust_rate"
                )

        return thrust_is_regulated, thrust_rate, thrust

    def sfc_at_max_thrust(
        self, atmosphere: Atmosphere, mach: Union[float, Sequence[float]]
    ) -> np.ndarray:
        """
        Computation of Specific Fuel Consumption at maximum thrust.

        Uses model described in :cite:`roux:2005`, p.41.

        :param atmosphere: Atmosphere instance at intended altitude
        :param mach: Mach number(s)
        :return: SFC (in kg/s/N)
        """

        altitude = atmosphere.get_altitude(False)
        mach = np.asarray(mach)

        # Following coefficients are constant for alt<=0 and alt >=11000m.
        # We use numpy to implement that so we are safe if altitude is a sequence.
        bound_altitude = np.minimum(11000, np.maximum(0, altitude))

        # pylint: disable=invalid-name  # coefficients are named after model
        a1 = -7.44e-13 * bound_altitude + 6.54e-7
        a2 = -3.32e-10 * bound_altitude + 8.54e-6
        b1 = -3.47e-11 * bound_altitude - 6.58e-7
        b2 = 4.23e-10 * bound_altitude + 1.32e-5
        c = -1.05e-7

        theta = atmosphere.temperature / ATM_SEA_LEVEL.temperature
        sfc = (
            mach * (a1 * self.bypass_ratio + a2)
            + (b1 * self.bypass_ratio + b2) * np.sqrt(theta)
            + ((7.4e-13 * (self.overall_pressure_ratio - 30) * altitude) + c)
            * (self.overall_pressure_ratio - 30)
        )

        return sfc

    def sfc_ratio(
        self,
        altitude: Union[float, Sequence[float]],
        thrust_rate: Union[float, Sequence[float]],
        mach: Union[float, Sequence[float]] = 0.8,
    ) -> np.ndarray:
        """
        Computation of ratio :math:`\\frac{SFC(F)}{SFC(Fmax)}`, given altitude
        and thrust_rate :math:`\\frac{F}{Fmax}`.

        Uses a patched version of model described in :cite:`roux:2002`, p.85.

        Warning: this model is very limited

        :param altitude:
        :param thrust_rate:
        :param mach: only used for logger checks as model is made for Mach~0.8
        :return: SFC ratio
        """

        altitude = np.asarray(altitude)
        thrust_rate = np.asarray(thrust_rate)
        mach = np.asarray(mach)

        delta_h = altitude - self.design_alt
        thrust_ratio_at_min_sfc_ratio = -9.6e-5 * delta_h + 0.85  # =Fi in model

        min_sfc_ratio = np.minimum(0.998, -3.385e-5 * delta_h + 0.995)

        # Get sfc_ratio_min closer to 1 when Fi closes to 1, to
        # respect coeff<=MAX_SFC_RATIO_COEFF
        # pylint: disable=unsupported-assignment-operation  # pylint is wrong
        min_sfc_ratio = np.maximum(
            min_sfc_ratio, 1 - MAX_SFC_RATIO_COEFF * (1 - thrust_ratio_at_min_sfc_ratio) ** 2
        )
        coeff = (1 - min_sfc_ratio) / (1 - thrust_ratio_at_min_sfc_ratio) ** 2

        # When thrust_ratio_at_min_sfc_ratio==1. (so min_sfc_ratio==1 also),
        # coeff has to be affected by hand
        if np.size(coeff) != 1:
            min_sfc_ratio[thrust_ratio_at_min_sfc_ratio == 1.0] = 1.0
            coeff[thrust_ratio_at_min_sfc_ratio == 1.0] = MAX_SFC_RATIO_COEFF
        elif thrust_ratio_at_min_sfc_ratio == 1.0:
            min_sfc_ratio = 1.0
            coeff = MAX_SFC_RATIO_COEFF

        return coeff * (thrust_rate - thrust_ratio_at_min_sfc_ratio) ** 2 + min_sfc_ratio

    def max_thrust(
        self,
        atmosphere: Atmosphere,
        mach: Union[float, Sequence[float]],
        delta_t4: Union[float, Sequence[float]],
    ) -> np.ndarray:
        """
        Computation of maximum thrust.

        Uses model described in :cite:`roux:2005`, p.57-58

        :param atmosphere: Atmosphere instance at intended altitude (should be <=20km)
        :param mach: Mach number(s) (should be between 0.05 and 1.0)
        :param delta_t4: (unit=K) difference between operational and design values of
                         turbine inlet temperature in K
        :return: maximum thrust (in N)
        """

        altitude = atmosphere.get_altitude(altitude_in_feet=False)
        mach = np.asarray(mach)
        delta_t4 = np.asarray(delta_t4)

        def _mach_effect():
            """Computation of Mach effect."""
            vect = [
                (self.overall_pressure_ratio - 30) ** 2,
                (self.overall_pressure_ratio - 30),
                1.0,
                self.t_4,
                delta_t4,
            ]

            def _calc_coef(a_coeffs, b_coeffs):
                # We don't use np.dot because delta_t4 can be a sequence
                return (
                    a_coeffs[0] * vect[0]
                    + a_coeffs[1] * vect[1]
                    + a_coeffs[2]
                    + a_coeffs[3] * vect[3]
                    + a_coeffs[4] * vect[4]
                ) * self.bypass_ratio + (
                    b_coeffs[0] * vect[0]
                    + b_coeffs[1] * vect[1]
                    + b_coeffs[2]
                    + b_coeffs[3] * vect[3]
                    + b_coeffs[4] * vect[4]
                )

            f_ms = _calc_coef(ALPHA[0], BETA[0])
            g_ms = _calc_coef(ALPHA[1], BETA[1])
            f_fm = _calc_coef(ALPHA[2], BETA[2])
            g_fm = _calc_coef(ALPHA[3], BETA[3])

            ms_11000 = (
                A_MS * self.t_4
                + B_MS * self.bypass_ratio
                + C_MS * (self.overall_pressure_ratio - 30)
                + D_MS * delta_t4
                + E_MS
            )

            fm_11000 = (
                A_FM * self.t_4
                + B_FM * self.bypass_ratio
                + C_FM * (self.overall_pressure_ratio - 30)
                + D_FM * delta_t4
                + E_FM
            )

            # Following coefficients are constant for alt >=11000m.
            # We use numpy to implement that so we are safe if altitude is a sequence.
            bound_altitude = np.minimum(11000, altitude)

            m_s = ms_11000 + f_ms * (bound_altitude - 11000) ** 2 + g_ms * (bound_altitude - 11000)
            f_m = fm_11000 + f_fm * (bound_altitude - 11000) ** 2 + g_fm * (bound_altitude - 11000)

            alpha_mach_effect = (1 - f_m) / (m_s * m_s)

            return alpha_mach_effect * (mach - m_s) ** 2 + f_m

        def _altitude_effect():
            """Computation of altitude effect."""
            # pylint: disable=invalid-name  # coefficients are named after model
            k = 1 + 1.2e-3 * delta_t4
            nf = 0.98 + 8e-4 * delta_t4

            def _troposhere_effect(density, altitude, k, nf):
                return (
                    k
                    * ((density / ATM_SEA_LEVEL.density) ** nf)
                    * (1 / (1 - (0.04 * np.sin((np.pi * altitude) / 11000))))
                )

            def _stratosphere_effect(density, k, nf):
                return (
                    k
                    * ((ATM_TROPOPAUSE.density / ATM_SEA_LEVEL.density) ** nf)
                    * density
                    / ATM_TROPOPAUSE.density
                )

            if np.size(altitude) == 1:
                if altitude <= 11000:
                    h = _troposhere_effect(atmosphere.density, altitude, k, nf)
                else:
                    h = _stratosphere_effect(atmosphere.density, k, nf)
            else:
                h = np.empty(np.shape(altitude))
                idx = altitude <= 11000
                if np.size(delta_t4) == 1:
                    h[idx] = _troposhere_effect(atmosphere.density[idx], altitude[idx], k, nf)
                    idx = np.logical_not(idx)
                    h[idx] = _stratosphere_effect(atmosphere.density[idx], k, nf)
                else:
                    h[idx] = _troposhere_effect(
                        atmosphere.density[idx], altitude[idx], k[idx], nf[idx]
                    )
                    idx = np.logical_not(idx)
                    h[idx] = _stratosphere_effect(atmosphere.density[idx], k[idx], nf[idx])

            return h

        def _residuals():
            """Computation of residuals."""
            return (
                -4.51e-3 * self.bypass_ratio
                + 2.19e-5 * self.t_4
                - 3.09e-4 * (self.overall_pressure_ratio - 30)
                + 0.945
            )

        return self.f_0 * _mach_effect() * _altitude_effect() * _residuals()

    def installed_weight(self) -> float:
        """
        Computes weight of installed engine, depending on MTO thrust (F0).

        Uses model described in :cite:`roux:2005`, p.74

        :return: installed weight (in kg)
        """
        # FIXME : separate raw engine weight and installation factor
        installation_factor = 1.2

        if self.f_0 < 80000:
            weight = 22.2e-3 * self.f_0
        else:
            weight = 14.1e-3 * self.f_0 + 648

        installed_weight = installation_factor * weight

        return installed_weight

    def length(self) -> float:
        # TODO: update model reference with last edition of Raymer
        """
        Computes engine length from MTO thrust and maximum Mach.

        Model from :cite:`raymer:1999`, p.74

        :return: engine length (in m)
        """
        length = 0.49 * (self.f_0 / 1000) ** 0.4 * self.mach_max**0.2

        return length

    def nacelle_diameter(self) -> float:
        # TODO: update model reference with last edition of Raymer
        """
        Computes nacelle diameter from MTO thrust and bypass ratio.

        Model of engine diameter from :cite:`raymer:1999`, p.235.
        Nacelle diameter is considered 10% greater (:cite:`kroo:2001`)

        :return: nacelle diameter (in m)
        """
        engine_diameter = 0.15 * (self.f_0 / 1000) ** 0.5 * math.exp(0.04 * self.bypass_ratio)
        nacelle_diameter = engine_diameter * 1.1
        return nacelle_diameter

    def _get_delta_t4(
        self, phase: Union[EngineSetting, Sequence[EngineSetting]]
    ) -> Union[float, Sequence[float]]:
        """
        :param phase:
        :return: DeltaT4 according to engine_setting
        """

        if np.shape(phase) == ():  # engine_setting is a scalar
            return self.dt4_values[phase]

        # Here engine_setting is a sequence. Ensure now it is a numpy array
        phase_array = np.asarray(phase)

        delta_t4 = np.empty(phase_array.shape)
        for phase_value, dt4_value in self.dt4_values.items():
            delta_t4[phase_array == phase_value] = dt4_value

        return delta_t4



"""Computation of CL characteristics at low speed."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2023 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_CL_ALPHA


@RegisterSubmodel(SERVICE_CL_ALPHA, "fastoad.submodel.aerodynamics.CLalpha.legacy")
class ComputeCLAlpha(om.ExplicitComponent):
    """
    Computes CL gradient.

    CL gradient from :cite:`raymer:1999` Eq 12.6
    """

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        if not self.options["low_speed_aero"]:
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_height", val=np.nan, units="m")
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")
        self.add_input("data:geometry:wing:aspect_ratio", val=np.nan)
        self.add_input("data:geometry:wing:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:sweep_25", val=np.nan, units="rad")
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:tip:thickness_ratio", val=np.nan)

        if self.options["low_speed_aero"]:
            self.add_output("data:aerodynamics:aircraft:low_speed:CL_alpha", units="1/rad")
        else:
            self.add_output("data:aerodynamics:aircraft:cruise:CL_alpha", units="1/rad")

    def setup_partials(self):
        if self.options["low_speed_aero"]:
            self.declare_partials("data:aerodynamics:aircraft:low_speed:CL_alpha", "*", method="fd")
        else:
            self.declare_partials("data:aerodynamics:aircraft:cruise:CL_alpha", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        width_max = inputs["data:geometry:fuselage:maximum_width"]
        height_max = inputs["data:geometry:fuselage:maximum_height"]
        span = inputs["data:geometry:wing:span"]
        lambda_wing = inputs["data:geometry:wing:aspect_ratio"]
        root_chord = inputs["data:geometry:wing:root:chord"]
        tip_chord = inputs["data:geometry:wing:tip:chord"]
        tip_thickness_ratio = inputs["data:geometry:wing:tip:thickness_ratio"]
        sweep_25 = inputs["data:geometry:wing:sweep_25"]
        wing_area = inputs["data:geometry:wing:area"]

        if self.options["low_speed_aero"]:
            mach = 0.2
        else:
            mach = inputs["data:TLAR:cruise_mach"]

        beta = np.sqrt(1.0 - mach**2)
        d_f = np.sqrt(width_max * height_max)
        fuselage_lift_factor = 1.07 * (1.0 + d_f / span) ** 2
        lambda_wing_eff = lambda_wing * (1.0 + 1.9 * tip_chord * tip_thickness_ratio / span)
        cl_alpha_wing = (
            2.0
            * np.pi
            * lambda_wing_eff
            / (
                2.0
                + np.sqrt(
                    4.0
                    + lambda_wing_eff**2
                    * beta**2
                    / 0.9025  # equals 0.95**2
                    * (1.0 + (np.tan(sweep_25)) ** 2 / beta**2)
                )
            )
            * (wing_area - root_chord * width_max)
            / wing_area
            * fuselage_lift_factor
        )

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:aircraft:low_speed:CL_alpha"] = cl_alpha_wing
        else:
            outputs["data:aerodynamics:aircraft:cruise:CL_alpha"] = cl_alpha_wing



"""
Estimation of wing center of gravity
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from ..constants import SERVICE_WING_CG


@RegisterSubmodel(SERVICE_WING_CG, "fastoad.submodel.weight.cg.wing.legacy")
class ComputeWingCG(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Wing center of gravity estimation"""

    def setup(self):
        self.add_input("data:geometry:wing:kink:span_ratio", val=np.nan)
        self.add_input("data:geometry:wing:spar_ratio:front:root", val=np.nan)
        self.add_input("data:geometry:wing:spar_ratio:front:kink", val=np.nan)
        self.add_input("data:geometry:wing:spar_ratio:front:tip", val=np.nan)
        self.add_input("data:geometry:wing:spar_ratio:rear:root", val=np.nan)
        self.add_input("data:geometry:wing:spar_ratio:rear:kink", val=np.nan)
        self.add_input("data:geometry:wing:spar_ratio:rear:tip", val=np.nan)
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")

        self.add_output("data:weight:airframe:wing:CG:x", units="m")

    def setup_partials(self):
        self.declare_partials("data:weight:airframe:wing:CG:x", "*", method="fd")

    def compute(self, inputs, outputs):
        wing_break = inputs["data:geometry:wing:kink:span_ratio"]
        front_spar_ratio_root = inputs["data:geometry:wing:spar_ratio:front:root"]
        front_spar_ratio_middle = inputs["data:geometry:wing:spar_ratio:front:kink"]
        front_spar_ratio_tip = inputs["data:geometry:wing:spar_ratio:front:tip"]
        rear_spar_ratio_root = inputs["data:geometry:wing:spar_ratio:rear:root"]
        rear_spar_ratio_middle = inputs["data:geometry:wing:spar_ratio:rear:kink"]
        rear_spar_ratio_tip = inputs["data:geometry:wing:spar_ratio:rear:tip"]
        span = inputs["data:geometry:wing:span"]
        x0_wing = inputs["data:geometry:wing:MAC:leading_edge:x:local"]
        l2_wing = inputs["data:geometry:wing:root:chord"]
        l3_wing = inputs["data:geometry:wing:kink:chord"]
        l4_wing = inputs["data:geometry:wing:tip:chord"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        x3_wing = inputs["data:geometry:wing:kink:leading_edge:x:local"]
        y3_wing = inputs["data:geometry:wing:kink:y"]
        y4_wing = inputs["data:geometry:wing:tip:y"]
        x4_wing = inputs["data:geometry:wing:tip:leading_edge:x:local"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]

        # TODO: make this constant an option
        if wing_break >= 0.35:
            y_cg = span / 2 * 0.35
            l_cg = (y3_wing - y_cg) / (y3_wing - y2_wing) * (l2_wing - l3_wing) + l3_wing
            front_spar_cg = (y3_wing - y_cg) / (y3_wing - y2_wing) * (
                l2_wing * front_spar_ratio_root - l3_wing * front_spar_ratio_middle
            ) + l3_wing * front_spar_ratio_middle
            rear_spar_cg = (y3_wing - y_cg) / (y3_wing - y2_wing) * (
                l2_wing * rear_spar_ratio_root - l3_wing * rear_spar_ratio_middle
            ) + l3_wing * rear_spar_ratio_middle
            x_cg = (
                (y_cg - y2_wing) / (y3_wing - y2_wing) * x3_wing
                + front_spar_cg
                + (l_cg - front_spar_cg - rear_spar_cg) * 0.7
            )
        elif wing_break < 0.35:
            y_cg = span / 2 * 0.35
            l_cg = (y4_wing - y_cg) / (y4_wing - y3_wing) * (l3_wing - l4_wing) + l4_wing
            front_spar_cg = (y4_wing - y_cg) / (y4_wing - y3_wing) * (
                l3_wing * front_spar_ratio_middle - l4_wing * front_spar_ratio_tip
            ) + l4_wing * front_spar_ratio_tip
            rear_spar_cg = (y4_wing - y_cg) / (y4_wing - y3_wing) * (
                l3_wing * rear_spar_ratio_middle - l4_wing * rear_spar_ratio_tip
            ) + l4_wing * rear_spar_ratio_tip

            x_cg = (
                (y_cg - y3_wing) / (y4_wing - y3_wing) * x4_wing
                + front_spar_cg
                + (l_cg - front_spar_cg - rear_spar_cg) * 0.7
            )
        x_cg_absolute = fa_length - 0.25 * x0_wing + (x_cg - x0_wing)

        outputs["data:weight:airframe:wing:CG:x"] = x_cg_absolute



"""
Estimation of landing gear weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_LANDING_GEARS_MASS


@RegisterSubmodel(
    SERVICE_LANDING_GEARS_MASS, "fastoad.submodel.weight.mass.airframe.landing_gears.legacy"
)
class LandingGearWeight(om.ExplicitComponent):
    """
    Weight estimation for landing gears

    Based on formulas in :cite:`supaero:2014`, mass contribution A5
    """

    def setup(self):
        self.add_input("data:weight:aircraft:MTOW", val=np.nan, units="kg")
        self.add_input("tuning:weight:airframe:landing_gear:mass:k", val=1.0)
        self.add_input("tuning:weight:airframe:landing_gear:mass:offset", val=0.0, units="kg")

        self.add_output("data:weight:airframe:landing_gear:main:mass", units="kg")
        self.add_output("data:weight:airframe:landing_gear:front:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mtow = inputs["data:weight:aircraft:MTOW"]
        k_a5 = inputs["tuning:weight:airframe:landing_gear:mass:k"]
        offset_a5 = inputs["tuning:weight:airframe:landing_gear:mass:offset"]

        temp_a51 = 18.1 + 0.131 * mtow**0.75 + 0.019 * mtow + 2.23e-5 * mtow**1.5
        temp_a52 = 9.1 + 0.082 * mtow**0.75 + 2.97e-6 * mtow**1.5

        a51 = k_a5 * temp_a51 + offset_a5
        a52 = k_a5 * temp_a52 + offset_a5

        outputs["data:weight:airframe:landing_gear:main:mass"] = a51
        outputs["data:weight:airframe:landing_gear:front:mass"] = a52



"""
Payload mass computation
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fastoad.module_management.service_registry import RegisterSubmodel
from openmdao import api as om

from .constants import SERVICE_PAYLOAD_MASS


@RegisterSubmodel(SERVICE_PAYLOAD_MASS, "fastoad.submodel.weight.mass.payload.legacy")
class ComputePayload(om.ExplicitComponent):
    """Computes payload from NPAX"""

    def setup(self):
        self.add_input("data:TLAR:NPAX", val=np.nan)
        self.add_input(
            "settings:weight:aircraft:payload:design_mass_per_passenger",
            val=90.72,
            units="kg",
            desc="Design value of mass per passenger",
        )
        self.add_input(
            "settings:weight:aircraft:payload:max_mass_per_passenger",
            val=130.72,
            units="kg",
            desc="Maximum value of mass per passenger",
        )

        self.add_output("data:weight:aircraft:payload", units="kg")
        self.add_output("data:weight:aircraft:max_payload", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        npax = inputs["data:TLAR:NPAX"]
        mass_per_pax = inputs["settings:weight:aircraft:payload:design_mass_per_passenger"]
        max_mass_per_pax = inputs["settings:weight:aircraft:payload:max_mass_per_passenger"]

        outputs["data:weight:aircraft:payload"] = npax * mass_per_pax
        outputs["data:weight:aircraft:max_payload"] = npax * max_mass_per_pax



"""
Estimation of pylons weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .constants import SERVICE_PYLONS_MASS


@RegisterSubmodel(SERVICE_PYLONS_MASS, "fastoad.submodel.weight.mass.airframe.pylons.legacy")
class PylonsWeight(om.ExplicitComponent):
    """
    Weight estimation for pylons

    Based on formula in :cite:`supaero:2014`, mass contribution A6
    """

    def setup(self):
        self.add_input("data:geometry:propulsion:pylon:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:propulsion:layout", val=np.nan)
        self.add_input("data:weight:propulsion:engine:mass", val=np.nan, units="kg")
        self.add_input("data:geometry:propulsion:engine:count", val=np.nan)
        self.add_input("tuning:weight:airframe:pylon:mass:k", val=1.0)
        self.add_input("tuning:weight:airframe:pylon:mass:offset", val=0.0, units="kg")

        self.add_output("data:weight:airframe:pylon:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        wet_area_pylon = inputs["data:geometry:propulsion:pylon:wetted_area"]
        weight_engine = inputs["data:weight:propulsion:engine:mass"]
        n_engines = inputs["data:geometry:propulsion:engine:count"]
        k_a6 = inputs["tuning:weight:airframe:pylon:mass:k"]
        offset_a6 = inputs["tuning:weight:airframe:pylon:mass:offset"]
        propulsion_layout = np.round(inputs["data:geometry:propulsion:layout"])

        if propulsion_layout == 1.0:
            temp_a6 = (
                1.2
                * wet_area_pylon**0.5
                * n_engines
                * (23 + 0.588 * (weight_engine / n_engines) ** 0.708)
            )
        else:
            temp_a6 = 0.08 * weight_engine

        outputs["data:weight:airframe:pylon:mass"] = k_a6 * temp_a6 + offset_a6



"""
Estimation of wing geometry
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import openmdao.api as om

from .constants import (
    SERVICE_WING_GEOMETRY_GLOBAL_POSITIONS,
    SERVICE_WING_GEOMETRY_MFW,
    SERVICE_WING_GEOMETRY_PLANFORM,
    SERVICE_WING_GEOMETRY_THICKNESS,
    SERVICE_WING_GEOMETRY_WET_AREA,
)
from ...constants import SERVICE_WING_GEOMETRY


@oad.RegisterSubmodel(SERVICE_WING_GEOMETRY, "fastoad.submodel.geometry.wing.legacy")
class ComputeWingGeometry(om.Group):
    """Wing geometry estimation"""

    def initialize(self):
        self.options.declare("compute_thicknesses", types=bool, default=True)

    def setup(self):
        self.add_subsystem(
            "planform",
            oad.RegisterSubmodel.get_submodel(SERVICE_WING_GEOMETRY_PLANFORM),
            promotes=["*"],
        )
        self.add_subsystem(
            "global_positions",
            oad.RegisterSubmodel.get_submodel(SERVICE_WING_GEOMETRY_GLOBAL_POSITIONS),
            promotes=["*"],
        )
        if self.options["compute_thicknesses"]:
            self.add_subsystem(
                "toc_wing",
                oad.RegisterSubmodel.get_submodel(SERVICE_WING_GEOMETRY_THICKNESS),
                promotes=["*"],
            )
        self.add_subsystem(
            "wetarea_wing",
            oad.RegisterSubmodel.get_submodel(SERVICE_WING_GEOMETRY_WET_AREA),
            promotes=["*"],
        )
        self.add_subsystem(
            "mfw",
            oad.RegisterSubmodel.get_submodel(SERVICE_WING_GEOMETRY_MFW),
            promotes=["*"],
        )



"""Convenience module for computing leading edge X positions of wing chords."""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om


class HTChordGlobalPositions(om.Group):
    """
    Computes leading edge X positions of horizontal tail chords (root, tip)
    with respect to aircraft nose.
    """

    def setup(self):
        self.add_subsystem("compute", ComputeChordGlobalPositions(), promotes=["*"])

        # AddSubtractComp does not allow to set default values of inputs and we do not
        # manage value discrepancies at global level, so we need to do it here, hence
        # the need for the current group.
        self.set_input_defaults(
            "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25", val=np.nan, units="m"
        )
        self.set_input_defaults(
            "data:geometry:horizontal_tail:tip:leading_edge:x:local", val=np.nan, units="m"
        )
        self.set_input_defaults(
            "data:geometry:horizontal_tail:MAC:leading_edge:x:local", val=np.nan, units="m"
        )
        self.set_input_defaults("data:geometry:horizontal_tail:MAC:length", val=np.nan, units="m")


class ComputeChordGlobalPositions(om.Group):
    """
    Computes leading edge X positions of horizontal tail chords with respect to aircraft nose.
    """

    def setup(self):
        self.add_subsystem(
            "MAC25_x",
            om.AddSubtractComp(
                "data:geometry:horizontal_tail:MAC:at25percent:x",
                [
                    "data:geometry:wing:MAC:at25percent:x",
                    "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25",
                ],
                scaling_factors=[1.0, 1.0],
                units="m",
            ),
            promotes=["*"],
        )
        self.add_subsystem(
            "MAC_x",
            om.AddSubtractComp(
                "data:geometry:horizontal_tail:MAC:leading_edge:x",
                [
                    "data:geometry:horizontal_tail:MAC:at25percent:x",
                    "data:geometry:horizontal_tail:MAC:length",
                ],
                scaling_factors=[1.0, -0.25],
                units="m",
            ),
            promotes=["*"],
        )
        self.add_subsystem(
            "center_x",
            om.AddSubtractComp(
                "data:geometry:horizontal_tail:center:leading_edge:x",
                [
                    "data:geometry:horizontal_tail:MAC:leading_edge:x",
                    "data:geometry:horizontal_tail:MAC:leading_edge:x:local",
                ],
                scaling_factors=[1.0, -1.0],
                units="m",
            ),
            promotes=["*"],
        )
        self.add_subsystem(
            "tip_x",
            om.AddSubtractComp(
                "data:geometry:horizontal_tail:tip:leading_edge:x",
                [
                    "data:geometry:horizontal_tail:center:leading_edge:x",
                    "data:geometry:horizontal_tail:tip:leading_edge:x:local",
                ],
                scaling_factors=[1.0, 1.0],
                units="m",
            ),
            promotes=["*"],
        )



"""
Estimation of horizontal tail sweeps
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import math

import numpy as np
import openmdao.api as om


class ComputeHTSweep(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Horizontal tail sweeps estimation"""

    def setup(self):
        self.add_input("data:geometry:horizontal_tail:center:chord", val=np.nan, units="m")
        self.add_input("data:geometry:horizontal_tail:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:horizontal_tail:span", val=np.nan, units="m")
        self.add_input("data:geometry:horizontal_tail:sweep_25", val=np.nan, units="rad")

        self.add_output("data:geometry:horizontal_tail:sweep_0", units="deg")
        self.add_output("data:geometry:horizontal_tail:sweep_100", units="deg")

    def setup_partials(self):
        self.declare_partials("data:geometry:horizontal_tail:sweep_0", "*", method="fd")
        self.declare_partials("data:geometry:horizontal_tail:sweep_100", "*", method="fd")

    def compute(self, inputs, outputs):
        b_h = inputs["data:geometry:horizontal_tail:span"]
        root_chord = inputs["data:geometry:horizontal_tail:center:chord"]
        tip_chord = inputs["data:geometry:horizontal_tail:tip:chord"]
        sweep_25_ht = inputs["data:geometry:horizontal_tail:sweep_25"]

        half_span = b_h / 2.0

        # The conversion to degrees is intentionally done here and not in OpenMDAO, so that
        # the outputs will naturally be displayed in degrees.
        sweep_0_ht = math.degrees(
            (math.atan(math.tan(sweep_25_ht) + 0.25 * (root_chord - tip_chord) / half_span))
        )
        sweep_100_ht = math.degrees(
            (math.atan(math.tan(sweep_25_ht) - 0.75 * (root_chord - tip_chord) / half_span))
        )

        outputs["data:geometry:horizontal_tail:sweep_0"] = sweep_0_ht
        outputs["data:geometry:horizontal_tail:sweep_100"] = sweep_100_ht



"""
Computation of lift and drag increment due to high-lift devices
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from importlib.resources import open_text

import numpy as np
import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel
from scipy import interpolate

from . import resources
from ..constants import SERVICE_HIGH_LIFT

LIFT_EFFECTIVENESS_FILENAME = "interpolation of lift effectiveness.txt"


@RegisterSubmodel(SERVICE_HIGH_LIFT, "fastoad.submodel.aerodynamics.high_lift.legacy")
class ComputeDeltaHighLift(om.ExplicitComponent):
    """
    Provides lift and drag increments due to high-lift devices
    """

    def initialize(self):
        self.options.declare("landing_flag", default=False, types=bool)

    def setup(self):
        if self.options["landing_flag"]:
            self.add_input("data:mission:sizing:landing:flap_angle", val=np.nan, units="deg")
            self.add_input("data:mission:sizing:landing:slat_angle", val=np.nan, units="deg")
            self.add_input("data:aerodynamics:aircraft:landing:mach", val=np.nan)
            self.add_output("data:aerodynamics:high_lift_devices:landing:CL")
            self.add_output("data:aerodynamics:high_lift_devices:landing:CD")
        else:
            self.add_input("data:mission:sizing:takeoff:flap_angle", val=np.nan, units="deg")
            self.add_input("data:mission:sizing:takeoff:slat_angle", val=np.nan, units="deg")
            self.add_input("data:aerodynamics:aircraft:takeoff:mach", val=np.nan)
            self.add_output("data:aerodynamics:high_lift_devices:takeoff:CL")
            self.add_output("data:aerodynamics:high_lift_devices:takeoff:CD")

        self.add_input("data:geometry:wing:sweep_0", val=np.nan, units="rad")
        self.add_input("data:geometry:wing:sweep_100_outer", val=np.nan, units="rad")
        self.add_input("data:geometry:flap:chord_ratio", val=np.nan)
        self.add_input("data:geometry:flap:span_ratio", val=np.nan)
        self.add_input("data:geometry:slat:chord_ratio", val=np.nan)
        self.add_input("data:geometry:slat:span_ratio", val=np.nan)
        self.add_input(
            "tuning:aerodynamics:high_lift_devices:landing:CD:multi_slotted_flap_effect:k",
            val=1.0,
        )
        self.add_input(
            "tuning:aerodynamics:high_lift_devices:landing:CL:multi_slotted_flap_effect:k",
            val=1.0,
        )

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        if self.options["landing_flag"]:
            flap_angle = inputs["data:mission:sizing:landing:flap_angle"]
            slat_angle = inputs["data:mission:sizing:landing:slat_angle"]
            mach = inputs["data:aerodynamics:aircraft:landing:mach"]
        else:
            flap_angle = inputs["data:mission:sizing:takeoff:flap_angle"]
            slat_angle = inputs["data:mission:sizing:takeoff:slat_angle"]
            mach = inputs["data:aerodynamics:aircraft:takeoff:mach"]

        k_cd_slot = inputs[
            "tuning:aerodynamics:high_lift_devices:landing:CD:multi_slotted_flap_effect:k"
        ]
        k_cl_slot = inputs[
            "tuning:aerodynamics:high_lift_devices:landing:CL:multi_slotted_flap_effect:k"
        ]
        le_sweep_angle = inputs["data:geometry:wing:sweep_0"]
        te_sweep_angle = inputs["data:geometry:wing:sweep_100_outer"]
        flap_chord_ratio = inputs["data:geometry:flap:chord_ratio"]
        flap_span_ratio = inputs["data:geometry:flap:span_ratio"]
        slat_chord_ratio = inputs["data:geometry:slat:chord_ratio"]
        slat_span_ratio = inputs["data:geometry:slat:span_ratio"]

        if self.options["landing_flag"]:
            outputs["data:aerodynamics:high_lift_devices:landing:CL"] = self._get_delta_cl(
                slat_angle,
                flap_angle,
                slat_span_ratio,
                flap_span_ratio,
                slat_chord_ratio,
                flap_chord_ratio,
                mach,
                le_sweep_angle,
                te_sweep_angle,
                k_cl_slot,
            )
            outputs["data:aerodynamics:high_lift_devices:landing:CD"] = self._get_delta_cd(
                slat_angle, flap_angle, slat_span_ratio, flap_span_ratio, k_cd_slot
            )
        else:
            outputs["data:aerodynamics:high_lift_devices:takeoff:CL"] = self._get_delta_cl(
                slat_angle,
                flap_angle,
                slat_span_ratio,
                flap_span_ratio,
                slat_chord_ratio,
                flap_chord_ratio,
                mach,
                le_sweep_angle,
                te_sweep_angle,
                k_cl_slot,
            )
            outputs["data:aerodynamics:high_lift_devices:takeoff:CD"] = self._get_delta_cd(
                slat_angle, flap_angle, slat_span_ratio, flap_span_ratio, k_cd_slot
            )

    def _get_delta_cl(
        self,
        slat_angle,
        flap_angle,
        slat_span_ratio,
        flap_span_ratio,
        slat_chord_ratio,
        flap_chord_ratio,
        mach,
        le_sweep_angle,
        te_sweep_angle,
        k_cl_slot,
    ):
        """
        Method based on Roskam book and Raymer book

        :param slat_angle: in degrees
        :param flap_angle: in degrees
        :param slat_span_ratio: the ratio of the part of wing which is equipped with slat
        :param flap_span_ratio: the ratio of the part of wing which is equipped with flap
        :param slat_chord_ratio: average ratio of slat chord to clean chord
        :param flap_chord_ratio: average ratio of flap chord to clean chord
        :param mach:
        :param le_sweep_angle: sweep angle at leading edge
        :param te_sweep_angle: sweep angle at trailing edge
        :param k_cl_slot: multiple slotted flap correction factor
        :return: increment of lift coefficient
        """

        flap_angle = np.radians(flap_angle)
        slat_angle = np.radians(slat_angle)

        #  ratio of chord with flap extended compared to clean chord
        ratio_c_flap = 1.0 + flap_chord_ratio * np.cos(flap_angle)

        alpha_flap = self._compute_alpha_flap(flap_angle * 57.3, flap_chord_ratio)

        # cl created by the flap in 2D
        delta_cl_flap = (
            2.0 * np.pi / np.sqrt(1 - mach**2) * ratio_c_flap * alpha_flap * flap_angle * k_cl_slot
        )

        # ratio of chord with slat extended compared to clean chord
        ratio_c_slat = 1.0 + slat_chord_ratio * np.cos(slat_angle)

        # leading edge slat effectiveness
        cl_delta = (
            5.05503e-7
            + 0.00666 * slat_chord_ratio
            + 0.23758 * slat_chord_ratio**2
            - 4.3639 * slat_chord_ratio**3
            + 51.16323 * slat_chord_ratio**4
            - 320.10803 * slat_chord_ratio**5
            + 1142.23033 * slat_chord_ratio**6
            - 2340.75209 * slat_chord_ratio**7
            + 2570.35947 * slat_chord_ratio**8
            - 1173.73465 * slat_chord_ratio**9
        )

        #  cl created by the slat in 2D
        delta_cl_slat = cl_delta * slat_angle * 57.3 * ratio_c_slat

        #  cl due to flap and slat including 3D
        delta_cl_total = delta_cl_flap * flap_span_ratio * np.cos(
            te_sweep_angle
        ) + delta_cl_slat * slat_span_ratio * np.cos(
            le_sweep_angle
        )  # this equation is from ref Raymer book

        return delta_cl_total

    def _get_delta_cd(self, slat_angle, flap_angle, slat_span_ratio, flap_span_ratio, k_cd_slot):
        """

        :param slat_angle: in degrees
        :param flap_angle: in degrees
        :param slat_span_ratio: the ratio of the part of wing which is equipped with slat
        :param flap_span_ratio: the ratio of the part of wing which is equipped with flap
        :param k_cd_slot: multiple slotted flap correction factor
        :return: increment of drag coefficient
        """

        cd0_slat = (
            (
                -0.00266
                + 0.06065 * slat_angle
                - 0.03023 * slat_angle**2
                + 0.01055 * slat_angle**3
                - 0.00176 * slat_angle**4
                + 1.77986e-4 * slat_angle**5
                - 1.11754e-5 * slat_angle**6
                + 4.19082e-7 * slat_angle**7
                - 8.53492e-9 * slat_angle**8
                + 7.24194e-11 * slat_angle**9
            )
            * slat_span_ratio
            / 100
        )
        cd0_flap = (
            (
                -0.01523
                + 0.05145 * flap_angle
                - 9.53201e-4 * flap_angle**2
                + 7.5972e-5 * flap_angle**3
            )
            * k_cd_slot
            * flap_span_ratio
            / 100
        )
        total_cd0 = cd0_flap + cd0_slat

        return total_cd0

    def _compute_alpha_flap(self, flap_angle, ratio_cf_flap):
        temp_array = []
        with open_text(resources, LIFT_EFFECTIVENESS_FILENAME) as fichier:
            for line in fichier:
                temp_array.append([float(x) for x in line.split(",")])
        x1 = []
        y1 = []
        x2 = []
        y2 = []
        x3 = []
        y3 = []
        x4 = []
        y4 = []
        x5 = []
        y5 = []

        for arr in temp_array:
            x1.append(arr[0])
            y1.append(arr[1])
            x2.append(arr[2])
            y2.append(arr[3])
            x3.append(arr[4])
            y3.append(arr[5])
            x4.append(arr[6])
            y4.append(arr[7])
            x5.append(arr[8])
            y5.append(arr[9])

        tck1 = interpolate.splrep(x1, y1, s=0)
        tck2 = interpolate.splrep(x2, y2, s=0)
        tck3 = interpolate.splrep(x3, y3, s=0)
        tck4 = interpolate.splrep(x4, y4, s=0)
        tck5 = interpolate.splrep(x5, y5, s=0)
        ynew1 = interpolate.splev(flap_angle, tck1, der=0)
        ynew2 = interpolate.splev(flap_angle, tck2, der=0)
        ynew3 = interpolate.splev(flap_angle, tck3, der=0)
        ynew4 = interpolate.splev(flap_angle, tck4, der=0)
        ynew5 = interpolate.splev(flap_angle, tck5, der=0)
        zs = [0.15, 0.20, 0.25, 0.30, 0.40]
        y_final = [ynew1, ynew2, ynew3, ynew4, ynew5]
        tck6 = interpolate.splrep(zs, y_final, s=0)
        return interpolate.splev(ratio_cf_flap, tck6, der=0)



"""
Main component for mass breakdown
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from openmdao.core.explicitcomponent import ExplicitComponent


class UpdateMLWandMZFW(ExplicitComponent):
    """
    Computes Maximum Landing Weight and Maximum Zero Fuel Weight from
    Overall Empty Weight and Maximum Payload.
    """

    def setup(self):
        self.add_input("data:weight:aircraft:OWE", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:max_payload", val=np.nan, units="kg")

        self.add_output("data:weight:aircraft:MZFW", units="kg")
        self.add_output("data:weight:aircraft:MLW", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        owe = inputs["data:weight:aircraft:OWE"][0]
        max_pl = inputs["data:weight:aircraft:max_payload"][0]
        mzfw = owe + max_pl
        mlw = 1.06 * mzfw

        outputs["data:weight:aircraft:MZFW"] = mzfw
        outputs["data:weight:aircraft:MLW"] = mlw



"""
Estimation of global center of gravity
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.module_management.service_registry import RegisterSubmodel

from .compute_max_cg_ratio import ComputeMaxCGratio
from ..constants import SERVICE_EMPTY_AIRCRAFT_CG, SERVICE_GLOBAL_CG, SERVICE_LOAD_CASES_CG


@RegisterSubmodel(SERVICE_GLOBAL_CG, "fastoad.submodel.weight.cg.global.legacy")
class ComputeGlobalCG(om.Group):
    # TODO: Document equations. Cite sources
    """Global center of gravity estimation"""

    def setup(self):
        self.add_subsystem(
            "cg_ratio_empty",
            RegisterSubmodel.get_submodel(SERVICE_EMPTY_AIRCRAFT_CG),
            promotes=["*"],
        )
        self.add_subsystem(
            "cg_ratio_load_cases",
            RegisterSubmodel.get_submodel(SERVICE_LOAD_CASES_CG),
            promotes=["*"],
        )
        self.add_subsystem("cg_ratio_max", ComputeMaxCGratio(), promotes=["*"])



"""
Weight computation (mass and CG)
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.module_management.constants import ModelDomain
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem, RegisterSubmodel

from .constants import SERVICE_CENTERS_OF_GRAVITY, SERVICE_MASS_BREAKDOWN
from ..constants import PAYLOAD_FROM_NPAX


@RegisterOpenMDAOSystem("fastoad.weight.legacy", domain=ModelDomain.WEIGHT)
class Weight(om.Group):
    """
    Computes masses and Centers of Gravity for each part of the empty operating aircraft, among
    these 5 categories:
    airframe, propulsion, systems, furniture, crew

    This model uses MTOW as an input, as it allows to size some elements, but resulting OWE do
    not aim at being consistent with MTOW.

    Consistency between OWE and MTOW can be achieved by cycling with a model that computes MTOW
    from OWE, which should come from a mission computation that will assess needed block fuel.
    """

    def initialize(self):
        self.options.declare(
            PAYLOAD_FROM_NPAX,
            types=bool,
            default=True,
            desc="If True (default), payload masses will be computed from NPAX.\n"
            "If False, design payload mass and maximum payload mass must be provided.",
        )

    def setup(self):
        self.add_subsystem(
            "cg", RegisterSubmodel.get_submodel(SERVICE_CENTERS_OF_GRAVITY), promotes=["*"]
        )
        self.add_subsystem(
            "mass_breakdown",
            RegisterSubmodel.get_submodel(
                SERVICE_MASS_BREAKDOWN, {PAYLOAD_FROM_NPAX: self.options[PAYLOAD_FROM_NPAX]}
            ),
            promotes=["*"],
        )



"""CG calculation for load cases."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om


class ComputeCGLoadCase(om.ExplicitComponent):
    """
    Base class for computing load cases for CG calculations.
    """

    @property
    def output_name(self):
        """Builds name of the unique output from option "case_number"."""
        return "data:weight:aircraft:load_case_%i:CG:MAC_position" % self.options["case_number"]

    def initialize(self):
        self.options.declare("case_number", 1, types=int)
        self.options.declare("mass_per_pax", 80.0, types=float)
        self.options.declare("mass_front_fret_per_pax", 0.0, types=float)
        self.options.declare("mass_rear_fret_per_pax", 0.0, types=float)
        self.options.declare("fuel_mass_variable", "", types=str, allow_none=True)

    def setup(self):
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:weight:payload:PAX:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:payload:rear_fret:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:payload:front_fret:CG:x", val=np.nan, units="m")
        self.add_input("data:TLAR:NPAX", val=np.nan)
        self.add_input("data:weight:aircraft_empty:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:aircraft_empty:mass", val=np.nan, units="kg")

        if self.options["fuel_mass_variable"]:
            self.add_input(self.options["fuel_mass_variable"], val=np.nan, units="kg")
            self.add_input("data:weight:fuel_tank:CG:x", val=np.nan, units="m")

        self.add_output(self.output_name)

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]
        x_cg_aircraft_empty = inputs["data:weight:aircraft_empty:CG:x"]
        mass_aircraft_empty = inputs["data:weight:aircraft_empty:mass"]

        mass_payload, x_cg_payload = self._get_payload_mass_and_cg(inputs)

        x_cg_aircraft_with_payload = (
            mass_aircraft_empty * x_cg_aircraft_empty + mass_payload * x_cg_payload
        ) / (mass_aircraft_empty + mass_payload)
        cg_ratio_aircraft_with_payload = (
            x_cg_aircraft_with_payload - fa_length + 0.25 * l0_wing
        ) / l0_wing
        outputs[self.output_name] = cg_ratio_aircraft_with_payload

    def _get_payload_mass_and_cg(self, inputs):
        """

        :param inputs:
        :return: tuple (payload mass, X-position of payload CG)
        """

        cg_pax = inputs["data:weight:payload:PAX:CG:x"]
        cg_rear_fret = inputs["data:weight:payload:rear_fret:CG:x"]
        cg_front_fret = inputs["data:weight:payload:front_fret:CG:x"]
        npax = inputs["data:TLAR:NPAX"]

        mass_pax = npax * self.options["mass_per_pax"]
        mass_front_fret = inputs["data:TLAR:NPAX"] * self.options["mass_front_fret_per_pax"]
        mass_rear_fret = inputs["data:TLAR:NPAX"] * self.options["mass_rear_fret_per_pax"]
        if self.options["fuel_mass_variable"]:
            mass_fuel = inputs[self.options["fuel_mass_variable"]]
            cg_fuel = inputs["data:weight:fuel_tank:CG:x"]
        else:
            mass_fuel = 0.0
            cg_fuel = 0.0

        mass_payload = mass_pax + mass_rear_fret + mass_front_fret + mass_fuel
        x_cg_payload = (
            mass_pax * cg_pax
            + mass_rear_fret * cg_rear_fret
            + mass_front_fret * cg_front_fret
            + mass_fuel * cg_fuel
        ) / mass_payload

        return mass_payload, x_cg_payload



"""Constants for mass submodels."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

SERVICE_OWE = "service.mass.owe"
SERVICE_PAYLOAD_MASS = "service.mass.payload"
SERVICE_AIRFRAME_MASS = "service.mass.airframe"
SERVICE_PROPULSION_MASS = "service.mass.propulsion"
SERVICE_SYSTEMS_MASS = "service.mass.systems"
SERVICE_FURNITURE_MASS = "service.mass.furniture"
SERVICE_CREW_MASS = "service.mass.crew"

SERVICE_GUST_LOADS = "service.gust_loads"



"""
Estimation of center of gravity for load case 2
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.module_management.service_registry import RegisterSubmodel

from .compute_cg_loadcase_base import ComputeCGLoadCase
from .compute_cg_loadcases import SERVICE_LOAD_CASE_CG_PREFIX

CASE_NUMBER = 2


@RegisterSubmodel(
    f"{SERVICE_LOAD_CASE_CG_PREFIX}.{CASE_NUMBER}",
    f"fastoad.submodel.weight.cg.load_case.legacy.{CASE_NUMBER}",
)
class ComputeCGLoadCase2(ComputeCGLoadCase):
    """Center of gravity estimation for load case 3"""

    def setup(self):
        self.options["case_number"] = CASE_NUMBER
        self.options["mass_per_pax"] = 90.0
        self.options["mass_front_fret_per_pax"] = 20.0
        self.options["mass_rear_fret_per_pax"] = 20.0
        self.options["fuel_mass_variable"] = "data:weight:aircraft:MFW"
        return super().setup()



"""Computation of Reynolds number"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2024 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fastoad.module_management.service_registry import RegisterSubmodel
from openmdao.core.explicitcomponent import ExplicitComponent
from stdatm import AtmosphereSI

from ..constants import SERVICE_REYNOLDS_COEFFICIENT


@RegisterSubmodel(
    SERVICE_REYNOLDS_COEFFICIENT,
    "fastoad.submodel.aerodynamics.reynolds_coefficient.legacy",
)
class ComputeReynolds(ExplicitComponent):
    """Computation of Reynolds number"""

    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        if self.options["low_speed_aero"]:
            self.add_input("data:aerodynamics:aircraft:takeoff:mach", val=np.nan)
            self.add_output("data:aerodynamics:wing:low_speed:reynolds")
        else:
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
            self.add_input("data:mission:sizing:main_route:cruise:altitude", val=np.nan, units="m")
            self.add_output("data:aerodynamics:wing:cruise:reynolds")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        if self.options["low_speed_aero"]:
            mach = inputs["data:aerodynamics:aircraft:takeoff:mach"]
            altitude = 0.0
        else:
            mach = inputs["data:TLAR:cruise_mach"]
            altitude = inputs["data:mission:sizing:main_route:cruise:altitude"]

        atm = AtmosphereSI(altitude)
        atm.mach = mach
        reynolds = atm.unitary_reynolds

        if self.options["low_speed_aero"]:
            outputs["data:aerodynamics:wing:low_speed:reynolds"] = reynolds
        else:
            outputs["data:aerodynamics:wing:cruise:reynolds"] = reynolds
